#include-once

;===================================================================================================
; desript.ion working library
; Version 2.30
; Support unicode descript.ion's
;===================================================================================================
; 2.30 - changed internal handling of line breaks from LF to CRLF
; 2.23 - added "modified" flag to descriptions
; 2.22 - fixed "\" duplication in single-line comments
;===================================================================================================

Global Const $_CP_ACP  = 0
Global Const $_CP_UTF8 = 65001
Global Enum $ENC_ANSI = 1, $ENC_UTF16, $ENC_UTF16BE, $ENC_UTF8
Global Enum $DI_FILENAME, $DI_DESCRIPTION, $DI_MODIFIED, $DI_MAX
Global Enum $DI_META_NAME, $DI_META_ENCODING, $DI_META_HASUNICHARS, $DI_META_HASMISSEDCHARS, $DI_META_MAX
Global Const $abEndMarker[4] = ["0x04C2", "0x0400C200", "0x000400C2", "0x04C382"]

Func _DescrIon_Load($sFileName)
    Local $sFile, $aFile
    Local $aResult[1][$DI_MAX] = [[0, 0, 0]] ; create empty result
    Local $nEncoding = 0 ; no file exists
    Local $bUniChars = False ; no unicode chars
    Local $bMissedChars = False

    If FileExists($sFileName) Then
        Local $hFile = FileOpen($sFileName, 16) ; binary

        ; Determine the encoding
        Local $bBOM1 = FileRead($hFile, 2)
        Switch $bBOM1
            Case "0xEFBB" ; UTF-8
                $bBOM2 = FileRead($hFile, 1)
                If $bBOM2 = "0xBF" Then 
                    $nEncoding = $ENC_UTF8
                Else
                    $nEncoding = 1 ; not quite correct, in this case file is broken
                EndIf
            Case "0xFFFE" ; UTF-16
                $nEncoding = $ENC_UTF16
            Case "0xFEFF" ; UTF-16 BE
                $nEncoding = $ENC_UTF16BE
            Case Else
                $nEncoding = $ENC_ANSI
                FileSetPos($hFile, 0, 0) ; ANSI
        EndSwitch
        
        $sFile = FileRead($hFile) ; read the rest
        FileClose($hFile)

        Local $sEndMarker = __GetMarkerString($nEncoding)

        Local $sDecoded = BinaryToString($sFile, $nEncoding)
        
        ; check in descript.ion has foreign unicode chars
        $bUniChars = $nEncoding = $ENC_ANSI ? False : _StringHasUniChars($sDecoded)

        ; set array of lines
        $aFile = StringSplit($sDecoded, @LF)

        Local $sLine, $nQuotePos, $sDescr, $nCount = 0, $sName

        ; define result array
        Dim $aResult[$aFile[0]+1][$DI_MAX] = [[$aFile[0], 0, 0]]

        For $i = 1 To $aFile[0]
            $sLine = StringStripCR($aFile[$i])
            If StringLeft($sLine, 1) = '"' Then
                $nQuotePos = StringInStr($sLine, '"', 0, 2)
                $sName = StringMid($sLine, 2, $nQuotePos - 2)
                $sDescr = StringTrimLeft($sLine, $nQuotePos + 1)
            Else
                $sName = StringLeft($sLine, StringInStr($sLine, " ") - 1)
                $sDescr = StringTrimLeft($sLine, StringInStr($sLine, " "))
            EndIf

            ; skip empty names (unlikely it may appear though)
            If $sName = "" Then ContinueLoop

            $nCount += 1

            If Not $bMissedChars And StringInStr($sName, "?") Then $bMissedChars = True

            $aResult[$nCount][0] = $sName
            $aResult[$nCount][1] = __ParseDescr($sDescr, $sEndMarker)
            $aResult[$nCount][2] = False
        Next

        ; in case empty lines was in file, redim array once
        If $nCount < $aResult[0][0] Then
            $aResult[0][0] = $nCount
            ReDim $aResult[$nCount + 1][$DI_MAX]
        EndIf
    Else
        SetExtended(1) ; ion created
    EndIf 
   
    ; set metadata (filename, encoding etc)
    Local $aMeta[$DI_META_MAX] = [$sFileName, $nEncoding, $bUniChars, $bMissedChars]
    $aResult[0][1] = $aMeta

    Return $aResult
EndFunc

Func _DescrIon_GetMeta(ByRef $aDescrIon, $nIndex)
    Local $aMeta = $aDescrIon[0][1]
    If IsArray($aMeta) And ($nIndex >= 0) And ($nIndex < UBound($aMeta)) Then Return $aMeta[$nIndex]
    Return SetError(1)
EndFunc

Func _DescrIon_SetMeta(ByRef $aDescrIon, $nIndex, $vValue)
    Local $aMeta = $aDescrIon[0][1]
    If IsArray($aMeta) And ($nIndex >= 0) And ($nIndex < UBound($aMeta)) Then 
        $aMeta[$nIndex] = $vValue
        $aDescrIon[0][1] = $aMeta
    Else
        Return SetError(1)
    EndIf
EndFunc

Func _DescrIon_ReplaceAnsiName(ByRef $aDescrIon, $sDescrFile)
    Local $sAnsiName
    For $i = 1 To $aDescrIon[0][0]
        If $aDescrIon[$i][0] == $sDescrFile Then Return 1
    Next
    $sAnsiName = __UnicodeToMultibyte($sDescrFile, $_CP_ACP, True)
    For $i = 1 To $aDescrIon[0][0]
        If $aDescrIon[$i][0] == $sAnsiName Then 
            $aDescrIon[$i][0] = $sDescrFile
            Return 2
        EndIf 
    Next

    Return SetError(1, 0, 0)
EndFunc

Func _DescrIon_GetDescr(ByRef $aDescrIon, $sDescrFile)
    Local $bCheck = _DescrIon_NeedNameCheck($aDescrIon)
    If $bCheck Then
        Local $sAnsiName = __UnicodeToMultibyte($sDescrFile, $_CP_ACP, True)
    EndIf
    For $i = 1 To $aDescrIon[0][0]
        If $aDescrIon[$i][0] = $sDescrFile Then Return $aDescrIon[$i][1]
        If $bCheck And ($aDescrIon[$i][0] == $sAnsiName) Then Return $aDescrIon[$i][1]
;        If $aDescrIon[$i][0] = $sDescrFile Then Return StringReplace($aDescrIon[$i][1], @LF, @CRLF)
;        If $bCheck And ($aDescrIon[$i][0] == $sAnsiName) Then Return StringReplace($aDescrIon[$i][1], @LF, @CRLF)
    Next
    Return SetError(1, 0, "")
EndFunc

Func _DescrIon_SetDescr(ByRef $aDescrIon, $sDescrFile, $sNewDescr)
    Local $bCheck = _DescrIon_NeedNameCheck($aDescrIon)
    If $bCheck Then
        Local $sAnsiName = __UnicodeToMultibyte($sDescrFile, $_CP_ACP, True)
    EndIf

;    $sNewDescr = StringReplace($sNewDescr, @CRLF, @LF)
;    $sNewDescr = StringReplace($sNewDescr, @CR, @LF)

    For $i = 1 To $aDescrIon[0][0]
        If $aDescrIon[$i][0] = $sDescrFile Then
            $aDescrIon[$i][1] = $sNewDescr
            $aDescrIon[$i][2] = True
            Return 1
        EndIf
        If $bCheck And ($aDescrIon[$i][0] == $sAnsiName) Then
            $aDescrIon[$i][1] = $sNewDescr
            $aDescrIon[$i][2] = True
            Return 1
        EndIf
    Next

    $aDescrIon[0][0] += 1
    ReDim $aDescrIon[$aDescrIon[0][0] + 1][$DI_MAX]
    $aDescrIon[$aDescrIon[0][0]][0] = $sDescrFile
    $aDescrIon[$aDescrIon[0][0]][1] = $sNewDescr
    $aDescrIon[$aDescrIon[0][0]][2] = True
EndFunc

Func _DescrIon_NeedNameCheck(ByRef $aDescrIon)
    Local $aMeta = $aDescrIon[0][1]
    Return ($aMeta[$DI_META_ENCODING] = $ENC_ANSI) And $aMeta[$DI_META_HASMISSEDCHARS]
EndFunc

Func _DescrIon_Save($sFileName, ByRef $aDescrIon)
    If Not IsArray($aDescrIon) Or ($aDescrIon[0][0] = 0) Then Return SetError(1)
    
    Local $bLF, $aMeta = $aDescrIon[0][1], $nCount = 0
    Local $nEncoding = $aMeta[$DI_META_ENCODING]
    Local $bCRLF = StringToBinary(@CRLF, $nEncoding)

    FileSetAttrib($sFileName, "-H")
    
    Local $hFile = FileOpen($sFileName, 2 + 16) ; $FO_OVERWRITE + $FO_BINARY
    
    ; writing BOMs
    Switch $nEncoding
        Case $ENC_UTF16
            FileWrite($hFile, "0xFFFE")
        Case $ENC_UTF16BE
            FileWrite($hFile, "0xFEFF")
        Case $ENC_UTF8
            FileWrite($hFile, "0xEFBBBF")
    EndSwitch
   
    For $i = 1 To $aDescrIon[0][0]
        If $aDescrIon[$i][0] = "" Then ContinueLoop ; skip files with empty names
        If $aDescrIon[$i][1] = "" Then ContinueLoop ; skip files with empty descriptions

        $nCount += 1

        $sName = $aDescrIon[$i][0]
        If StringInStr($sName, " ") Then $sName = '"' & $sName & '"'
        $sLine = $aDescrIon[$i][1]
        $bLF = StringInStr($sLine, @CRLF) > 0 ; is multiline?
        If $bLF Then
            $sLine = StringReplace($sLine, "\", "\\")
            $sLine = StringReplace($sLine, @CRLF, "\n")
            $sLine = StringReplace($sLine, @LF, "\n") ; unlikely
        EndIf
        $aDescrIon[$i][$DI_MODIFIED] = False
        
        ; write line with filename + description        
        __FileWriteEnc($hFile, $sName & " " & $sLine, $nEncoding)
        ; write multiline end marker only in case of multiline comment
        If $bLF Then FileWrite($hFile, __GetMarker($nEncoding))
        ; write CRLF, need to be correctly encoded too!
        FileWrite($hFile, $bCRLF)
    Next

    FileClose($hFile)

    If $nCount = 0 Then
        FileDelete($sFileName)
    Else
        FileSetAttrib($sFileName, "+H")
    EndIf
EndFunc

; ************************************************************************
; Internal helper functions
; ************************************************************************

; Return multiline end marker by encoding
Func __GetMarker($nEnc)
    Return $abEndMarker[$nEnc-1]
EndFunc

; Return multiline end marker as string by encoding
Func __GetMarkerString($nEnc)
    Return BinaryToString($abEndMarker[$nEnc-1], $nEnc)            
EndFunc

Func __ParseDescr($sDescr, $sEndMarker)
    $sDescr = StringReplace($sDescr, $sEndMarker, "")
    If @extended = 0 Then Return $sDescr
    Return StringReplace(StringFormat($sDescr), @LF, @CRLF) ; a bit simpler then legacy code...
EndFunc

; write string encoded with given encoding
Func __FileWriteEnc($hFile, $vData, $nEnc)
    Local $bData = StringToBinary($vData, $nEnc)
    FileWrite($hFile, $bData)
EndFunc

Func _StringHasUniChars($vUnicode, $iCodePage = 0)
    If $vUnicode = "" Then Return False
    Local Const $WC_NO_BEST_FIT_CHARS = 0x00000400
	Local $aResult = DllCall("kernel32.dll", "int", "WideCharToMultiByte", _
                                             "uint", $iCodePage, _
                                             "dword", $WC_NO_BEST_FIT_CHARS, _
                                             "wstr", $vUnicode, _
                                             "int", -1, _
                                             "ptr", 0, _
                                             "int", 0, _
                                             "ptr", 0, _
                                             "bool*", 0)
	If @error Or Not $aResult[0] Then Return SetError(@error, @extended, false)
    Return $aResult[8] <> 0
EndFunc

Func __UnicodeToMultibyte($vUnicode, $iCodePage = 0, $bRetString = False)

	Local $aResult = DllCall("kernel32.dll", "int", "WideCharToMultiByte", _
                                             "uint", $iCodePage, _
                                             "dword", 0, _
                                             "wstr", $vUnicode, _
                                             "int", -1, _
                                             "ptr", 0, _
                                             "int", 0, _
                                             "ptr", 0, _
                                             "ptr", 0)
	If @error Or Not $aResult[0] Then Return SetError(@error + 20, @extended, "")

    Local $nAnsiSize = $aResult[0]
	Local $tMultiByte = DllStructCreate("byte[" & $nAnsiSize & "]")

	$aResult = DllCall("kernel32.dll", "int", "WideCharToMultiByte", _
                                       "uint", $iCodePage, _
                                       "dword", 0, _
                                       "wstr", $vUnicode, _
                                       "int", -1, _
                                       "struct*", $tMultiByte, _
                                       "int", $nAnsiSize, _
                                       "ptr", 0, _
                                       "ptr", 0)
	If @error Or Not $aResult[0] Then Return SetError(@error + 10, @extended, "")

	If $bRetString Then 
        Local $tAnsi = DllStructCreate("char[" & $nAnsiSize & "]", DllStructGetPtr($tMultiByte))
        Return DllStructGetData($tAnsi, 1)
    EndIf
	Return $tMultiByte
EndFunc