#include-once

; relatively rough thing to correct gui size on the 120 and higher dpi screens
; need Opt("GUIResizeMode", $GUI_DOCKAUTO) to be set before gui create
; also this functon should be placed AFTER GUISetState(@SW_SHOW)
Func _GUI_AdjustSize($hWnd)
    Local $aDPI = _GUI_GetDPI()
    If $aDPI[0] = 96 Then Return
    Local $iDPIRat = $aDPI[1]
    Local $pos = WinGetPos($hWnd)
    Local $relX =  (($pos[2] * $iDPIRat) - $pos[2]) / 2
    Local $relY =  (($pos[3] * $iDPIRat) - $pos[3]) / 2
    WinMove($hWnd, "", $pos[0] - $relX, $pos[1] - $relY, $pos[2] * $iDPIRat, $pos[3] * $iDPIRat)
EndFunc

; Set DPI awareness for HiDPI modes, this should be called once before _GUI_AdjustSize is called for GUI
; 1 - PROCESS_SYSTEM_DPI_AWARE, 2 - PROCESS_PER_MONITOR_DPI_AWARE (if app can reflect DPI changing)
Func _GUI_SetDPIAware($nMode = 1)
    Local $aRet = DllCall("Shcore.dll", "int", "SetProcessDpiAwareness", "int", $nMode)
    Return (Not @error) And ($aRet[0] = 0 Or $aRet[0] = 0x80000009) ; True if function exist and result success or already set up
EndFunc

Func _GUI_MenuRightJustify($hGUI, $nItem, $isIndex = false)
    Local $MFT_RIGHTJUSTIFY = 0x4000
    Local $MIIM_TYPE = 0x10

    Local $tagMENUITEMINFO = "uint;uint;uint;uint;uint;uint;uint;uint;dword;ptr;uint"
    Local $menu = DllCall("user32.dll", "hwnd", "GetMenu", "hwnd", $hGUI)
    $menu = $menu[0]

    Local $menuinfo = DllStructCreate($tagMENUITEMINFO)
    Local $buffer   = DllStructCreate("byte[80]")

    DllStructSetData($menuinfo, 1, DllStructGetSize($menuinfo))
    DllStructSetData($menuinfo, 2, $MIIM_TYPE)
    DllStructSetData($menuinfo, 10, DllStructGetPtr($buffer))
    DllStructSetData($menuinfo, 11, 80)

    DllCall("user32.dll", "int", "GetMenuItemInfo", "int", $menu, _
                                                    "int", $nItem, _
                                                    "int", $isIndex, _
                                                    "ptr", DllStructGetPtr($menuinfo))

    DllStructSetData($menuinfo, 3, BitOR(DllStructGetData($menuinfo, 3), $MFT_RIGHTJUSTIFY))

    DllCall("user32.dll", "int", "SetMenuItemInfo", "int", $menu, _
                                                    "int", $nItem, _
                                                    "int", $isIndex, _
                                                    "ptr", DllStructGetPtr($menuinfo))
EndFunc

Func _GUI_CenterToParent($hParent, $hWnd)
    Local $aParentPos = WinGetPos($hParent)
    If @error Then Return SetError(1)
    Local $aPos = WinGetPos($hWnd)
    Local $nX = $aParentPos[0] + $aParentPos[2] * 0.5 - $aPos[2] * 0.5
    Local $nY = $aParentPos[1] + $aParentPos[3] * 0.5 - $aPos[3] * 0.5
    WinMove($hWnd, "", $nX, $nY)
EndFunc

Func _GUI_GetDPI()
    ; Modified by YDY
    ; Author: Phillip123Adams
    Local $aRet[2], $iDPI, $aCall
    Local $iDPIRat, $hWnd = 0
    Local $hDC = DllCall("user32.dll", "long", "GetDC", "long", $hWnd)[0]
    $aCall = DllCall("gdi32.dll", "long", "GetDeviceCaps", "long", $hDC, "long", 90) ; LOGPIXELSY
    DllCall("user32.dll", "long", "ReleaseDC", "long", $hWnd, "long", $hDC)
    $iDPI = $aCall[0]
    ; Set a ratio for the GUI dimensions based upon the current DPI value.
    Select
      Case $iDPI = 0
         $iDPI = 96
         $iDPIRat =  94
      Case $iDPI < 84
         $iDPIRat =  $iDPI / 105
      Case $iDPI < 121
         $iDPIRat =  $iDPI / 96
      Case $iDPI < 145
         $iDPIRat =  $iDPI / 95
      Case Else
         $iDPIRat =  $iDPI / 94
    EndSelect
    $aRet[0] = $iDPI
    $aRet[1] = $iDPIRat
    Return $aRet
EndFunc

Func _GUI_TrackPopupMenu($hWnd, $hMenu, $x = Default, $y = Default, $nFlags = 0)
    If $x = Default Then $x = MouseGetPos()[0]
    If $y = Default Then $y = MouseGetPos()[1]
    Return _TrackPopupMenu($hWnd, $hMenu, $x, $y, $nFlags)
EndFunc

; simple helper to pass keep handles without needing to create globals (would be a lot faster with maps)
Func _GUI_Handles($vID, $vValue = Default)
    Local Static $aHandles[0][2]
    Local $iBound = UBound($aHandles)
    For $i = 0 To $iBound - 1
        If $aHandles[$i][0] = $vID And $vValue = Default Then Return $aHandles[$i][1]
    Next
    ReDim $aHandles[$iBound + 1][2]
    $aHandles[$iBound][0] = $vID
    $aHandles[$iBound][1] = $vValue
    Return $vValue
EndFunc

Func _GUI_NotifyWindow($hParent, $sText, $nWidth, $nHeight, $nDelay = 1000)
    Local $nMsg
    Local $nMargin = 15
    Local $hNotifyWindow = GUICreate("", $nWidth, $nHeight, -1, -1, 0x80880000, 0x00010000, $hParent) ; $WS_POPUPWINDOW, $WS_BORDER ; $WS_EX_CONTROLPARENT
    GUISetBkColor(DllCall("user32.dll", "dword", "GetSysColor", "int", 5)[0])
    Local $hText = GUICtrlCreateLabel($sText, $nMargin, $nMargin, $nWidth - 2 * $nMargin, $nHeight - 2 * $nMargin, 0x00002001) ; SS_CENTER, SS_EDITCONTROL

    _GUI_CenterToParent($hParent, $hNotifyWindow)

    GUISetState()
    GUISetState(@SW_DISABLE, $hParent)

    Sleep($nDelay)

    GUISetState(@SW_ENABLE, $hParent)
    GUIDelete($hNotifyWindow)
EndFunc

; ==================================================================================================
; Wrappers
; ==================================================================================================

; Show Popup menu at given coordinate
Func _TrackPopupMenu($hWnd, $hMenu, $x, $y, $nFlags = 0)
    Local $aRet = DllCall("user32.dll", "int", "TrackPopupMenuEx", "hwnd", $hMenu, "int", $nFlags, "int", $x, "int", $y, "hwnd", $hWnd, "ptr", 0)
    Return $aRet[0]
EndFunc

; Convert the client (GUI) coordinates to screen (desktop) coordinates
Func _ClientToScreen($hWnd, ByRef $x, ByRef $y)
    Local $stPoint = DllStructCreate("int;int")
    
    DllStructSetData($stPoint, 1, $x)
    DllStructSetData($stPoint, 2, $y)

    DllCall("user32.dll", "int", "ClientToScreen", "hwnd", $hWnd, "ptr", DllStructGetPtr($stPoint))
    
    $x = DllStructGetData($stPoint, 1)
    $y = DllStructGetData($stPoint, 2)
EndFunc

Func _ScreenToClient($hWnd, ByRef $x, ByRef $y)
    Local $stPoint = DllStructCreate("int;int")
    
    DllStructSetData($stPoint, 1, $x)
    DllStructSetData($stPoint, 2, $y)

    DllCall("user32.dll", "int", "ScreenToClient", "hwnd", $hWnd, "ptr", DllStructGetPtr($stPoint))
    
    $x = DllStructGetData($stPoint, 1)
    $y = DllStructGetData($stPoint, 2)
EndFunc
