#cs =============================================================================================================================
[info]
Param=%WF "%P"
Hotkey=Ctrl+Z
Info.ENG.Title=Comment editing
Info.ENG.Description=Editing comments for multiple files at once
Info.RUS.Title= 
Info.RUS.Description=     
Info.META.Version=3.2.0
Info.META.Date=2023-07-
Info.META.Author=Dmitry Yudin
[/info]
; Icon from "Oxygen Icon" set, oxygen-icons.org
#ce =============================================================================================================================

#pragma compile(FileDescription, Multi Comment Edit)
#pragma compile(FileVersion, 3.2.0)
#pragma compile(LegalCopyright, 2010-2023 Dmitry Yudin)
#pragma compile(Icon, tc_multi_comment_edit.ico)

; custom includes
#include <TC\TCHelpers.au3>
#include <TC\TCWDXPlugin.au3>
#include <Common\GUIMsgHandler.au3>
#include <Common\GUIHelpers.au3>
#include <Common\DescrIon.au3>
#include <Common\EmbedIcon.au3>

; stadard includes
#include <ButtonConstants.au3>
#include <EditConstants.au3>
#include <ComboConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>
#include <WindowsConstants.au3>
#include <ListViewConstants.au3>
#include <ListBoxConstants.au3>

#NoTrayIcon

If $CmdLine[0] = 0 Then 
    ; handling updir
    Global $g_sCurrPath = _TryGetBasePath()
    If @error then _TCHCheckCmdLine(2)
    $g_sCurrPath = _TCHEnsureSlash($g_sCurrPath, False)
     ; add current dir to faked list
    Global $g_aList[1] = [_TCHExtractFileName($g_sCurrPath)]
    ; no actual file, only used for checking unicode chars
    Global $g_sListFile = $g_aList[0]
    ; get parent path
    $g_sCurrPath = _TCHExtractFilePath($g_sCurrPath)
Else
    Global $g_sListFile = StringStripWS(FileRead($CmdLine[1]), 2)
    Global $g_aList = StringSplit($g_sListFile, @CRLF, 3)
    Global $g_sCurrPath = $CmdLine[2]
EndIf

Global $g_sFlags = $CmdLine[0] >= 3 ? $CmdLine[3] : ""

Global $g_sScriptConfig = _TCHChangeFileExt(@ScriptFullPath, "ini")
Global $bMultiOnly = IniRead($g_sScriptConfig, "Main", "MultiOnly", 0) = 1

; Only single file is selected and MCE_MULTI_ONLY set (or MultiOnly=1), invoke standard function
$bMultiOnly = $bMultiOnly Or StringInStr($g_sFlags, "MCE_MULTI_ONLY") > 0
If (UBound($g_aList) < 2) And $bMultiOnly And Not ($g_aList[0] = "..")  Then
    _TCHPostMsg(2700) ; Post "cm_EditComment" message to active TC instance
    Exit
EndIf

; some common constants, must be commented in case of using standard libraries
Global Const $SS_EDITCONTROL = 0x00002000
Global Const $TPM_RIGHTALIGN = 0x00000008

OnAutoItExitRegister("_OnExit")

; folders should not have "\" at the end
For $i = 0 To UBound($g_aList)-1
    $g_aList[$i] = _TCHEnsureSlash($g_aList[$i], False)
Next

Global $g_sCurrFile = ""

Global $g_sDescrIon = $g_sCurrPath & "descript.ion"

Global $nDefEnc            = IniRead($g_sScriptConfig, "Main", "DefEncoding", $ENC_UTF8)
Global $bPreferANSI        = IniRead($g_sScriptConfig, "Main", "PreferANSI", 0) = 1
Global $bDisableSingleEdit = IniRead($g_sScriptConfig, "Main", "DisableSingleEdit", 0) = 1

If $nDefEnc < $ENC_UTF16 And $nDefEnc > $ENC_UTF8 Then $nDefEnc = $ENC_UTF8

; handle old style flags
If $g_sFlags <> "" Then
    Global $aFlags = StringSplit($CmdLine[3], "|")
    For $i = 1 To $aFlags[0]
        Switch $aFlags[$i]
            Case "MCE_DEFENC_UTF8"
                $nDefEnc = $ENC_UTF8
            Case "MCE_DEFENC_UTF16"
                $nDefEnc = $ENC_UTF16
            Case "MCE_DEFENC_UTF16BE"
                $nDefEnc = $ENC_UTF16BE
            Case "MCE_DISABLE_SINGLEEDIT"
                $bDisableSingleEdit = True
            Case "MCE_PREFER_ANSI"
                $bPreferANSI = True
        EndSwitch
    Next
EndIf

; =========================================================================
Global $g_sMCEPluginName = "MCE"
; plugins array var, initialized later, only if required
Global $aPlugins = 0
; use non-printing separator char in GUI* functions to allow use "|" in templates
Global Const $cSep = Chr(0x1)
Opt("GUIDataSeparatorChar", $cSep)
; =========================================================================

$LANG("1",  "File Comment")
$LANG("2",  "Comments for %d files")
$LANG("3",  "Options")
$LANG("4",  "Replace")
$LANG("5",  "Add to beginning")
$LANG("6",  "Add to end")
$LANG("7",  "Cancel")
$LANG("8",  "Save")
$LANG("9",  "Apply")
$LANG("10", "Multiple files selected")
$LANG("11", "Edit comment for:")
$LANG("12", "Can't create comment on FTP.")
$LANG("13", "Can't create comment in plugin.")
$LANG("14", "Base path is empty, can't create comment for root dir.")
$LANG("15", "Encoding:")
$LANG("16", "Help")
$LANG("17", "Changing encoding from unicode to ANSI may cause data loss!")
$LANG("18", "When changing from ANSI to unicode, file names may need to be converted. Do you want the program try to do this?")
$LANG("19", "None")
$LANG("20", "Insert")
$LANG("21", "Plugins")
$LANG("22", "Templates")
$LANG("23", "Neighbour")
$LANG("24", "Remove template")
$LANG("25", "Replace macro")
$LANG("26", "Save text as template")
$LANG("27", "Do you want to remove this template?")
$LANG("28", "Template saved successfully.") ; redone
$LANG("29", "Macro")
$LANG("30", "Value")
$LANG("31", "Multiedit mode")
$LANG("32", "Single edit mode")
$LANG(LangInit)

; =========================================================================

; Check correctness of base path (not all variants can be checked...)
Select
    ; FTP
    Case StringRegExp($g_sCurrPath, "^/.+")
        MsgBox(48, "", $LANG("12"))
        Exit
    ; Inside plugin, but can't check the case, when create comment on plugin or any other item in network pane
    Case StringRegExp($g_sCurrPath, "^\\{3}[^\\]*") 
        MsgBox(48, "", $LANG("13"))
        Exit
    ; Double dots ".." in root dir when this option enabled
    Case $g_sCurrPath = ""
        MsgBox(48, "", $LANG("14"))
        Exit
EndSelect

Global $g_nCurEnc, $g_nNewEnc

Global $aDescr = _DescrIon_Load($g_sDescrIon)
If @extended = 1 Then ; created new
    $g_nCurEnc = 0
    $g_nNewEnc = $bPreferANSI ? $ENC_ANSI : $nDefEnc
Else
    $g_nCurEnc = _DescrIon_GetMeta($aDescr, $DI_META_ENCODING)
    $g_nNewEnc = $g_nCurEnc
EndIf

; global flag, allow to swith back to ANSI after removing unicode chars
Global $g_bFallbackANSI = $g_nNewEnc = $ENC_ANSI

; global flag, true if names are unicode...
Global $g_bNeedUnicode = _StringHasUniChars($g_sListFile)

; ... OR existing descript.ion has unicode chars
If Not $g_bNeedUnicode Then $g_bNeedUnicode = _DescrIon_GetMeta($aDescr, $DI_META_HASUNICHARS)

_UpdateEncState($g_bNeedUnicode)

Opt("GUIResizeMode", $GUI_DOCKAUTO)

; disable total commander window, cause a problem if the script is crashed :)
_TCHSetParentModal($TC_HWND, True)

; Set DPI awareness for HiDPI modes
_GUI_SetDPIAware()

; run main dialog
_MainDialog($TC_HWND)

; ==================================================================================================
; Main dialog functions
; ==================================================================================================
Func _MainDialog($hParent)
    Local $bModified = False
    Local $bMultiFile = UBound($g_aList) > 1

    Local $hGUI = GUICreate($LANG("1"), 625, 313, -1, -1, BitOR($WS_POPUP, $WS_CAPTION, $WS_SYSMENU), $WS_EX_COMPOSITED + $WS_EX_CONTEXTHELP, $hParent)

    ; Options popup menu
    Local $hOptMenu = GUICtrlCreateContextMenu(GUICtrlCreateDummy())
    Local $ahOptions[3]
    $ahOptions[0] = GUICtrlCreateMenuItem($LANG("4"), $hOptMenu, -1 , 1)
    $ahOptions[1] = GUICtrlCreateMenuItem($LANG("6"), $hOptMenu, -1 , 1)
    $ahOptions[2] = GUICtrlCreateMenuItem($LANG("5"), $hOptMenu, -1 , 1)
    GUICtrlCreateMenuItem("", $hOptMenu)
    Local $hMacro = GUICtrlCreateMenuItem($LANG("25"), $hOptMenu)
    GUICtrlSetState(-1, $GUI_CHECKED)

    ; OK button popup menu
    Local $hOKMenu = GUICtrlCreateContextMenu(GUICtrlCreateDummy())
    Local $hApply = GUICtrlCreateMenuItem($LANG("9") & @TAB & "Alt+Enter", $hOKMenu)
    Local $hSave = GUICtrlCreateMenuItem($LANG("8")& @TAB & "Ctrl+S", $hOKMenu)
    GUICtrlSetState(-1, $GUI_DISABLE)
    GUICtrlCreateMenuItem("", $hOKMenu)
    Local $hTplSave = GUICtrlCreateMenuItem($LANG("26")& @TAB & "Alt+T", $hOKMenu)

    ; encoding group
    GUICtrlCreateLabel($LANG("15"), 404, 8, 70, 17, $SS_CENTER)
    Local $hPrevEnc = GUICtrlCreateLabel(_GetEncString($g_nCurEnc), 474, 8, 50, 17, $SS_CENTER)
    GUICtrlSetFont(-1, -1, 800)
    GUICtrlCreateLabel(">>", 524, 8, 16, 17)
    Global $hEncSwitch = GUICtrlCreateCombo("", 540, 4, 80, 24, BitOR($CBS_DROPDOWNLIST, $CBS_AUTOHSCROLL))
    GUICtrlSetData($hEncSwitch, _GetEncComboString($g_bNeedUnicode))

    ; Files combo, main edit
    Local $hMultiEdit = GUICtrlCreateCheckbox($LANG("30"), 4, 5, 23, 23, BitOR($BS_PUSHLIKE, $BS_NOTIFY, $BS_ICON))
    GUICtrlSetState($hMultiEdit, $GUI_CHECKED)

    If $bMultiFile Then
        Local $hEdit = GUICtrlCreateEdit("", 4, 32, 616, 244, BitOR($ES_WANTRETURN,$WS_VSCROLL))
        If $bDisableSingleEdit Then
            $hEditComm = GUICtrlCreateLabel(StringFormat($LANG("2"), UBound($g_aList)), 8, 8, 400, 17, $DT_PATH_ELLIPSIS)
            GUICtrlSetState($hMultiEdit, $GUI_HIDE)
        Else
            Local $hEditComm = GUICtrlCreateCombo("", 30, 6, 366, 17, BitOR($CBS_DROPDOWNLIST, $WS_VSCROLL))
            GUICtrlSetData($hEditComm, _FileListToString($g_aList))
            GUICtrlSendMsg($hEditComm, $CB_SETCURSEL , 0, 0)
        EndIf
    Else
        $g_sCurrFile = $g_aList[0]
        Local $hEdit = GUICtrlCreateEdit("", 4, 40, 616, 236, BitOR($ES_WANTRETURN,$WS_VSCROLL))
        GUICtrlCreateLabel($LANG("11"), 8, 4, 296, 17, 0)
        Local $hEditComm = GUICtrlCreateLabel($g_sCurrFile, 8, 22, 400, 17, $DT_PATH_ELLIPSIS)
        GUICtrlSetTip($hEditComm, $g_sCurrFile)
        GUICtrlSetState($hMultiEdit, $GUI_HIDE)
    EndIf

    _GUICtrlSetCustomFont($hEdit, "EditorFont")

    ; edit handle is used quite frequently 
    Local $hEditWnd = GUICtrlGetHandle($hEdit)

    Local $hSize = GUICtrlCreateLabel("", 269, 289, 84, 17, $SS_CENTER)

    Local $hOptions = GUICtrlCreateButton($LANG("3"), 8, 283, 75, 24, $BS_NOTIFY)
    Local $hInsert  = GUICtrlCreateButton($LANG("20"), 90, 283, 75, 24)
    Local $hOK      = GUICtrlCreateButton("F2 OK", 448, 283, 80, 24, $BS_SPLITBUTTON) ; $BS_DEFPUSHBUTTON
    Local $hCancel  = GUICtrlCreateButton($LANG("7"), 536, 283, 80, 24, 0)

    ; assign main accelerators
    Local $hMultiEditHK = GUICtrlCreateDummy()

    Local $hGUI_AccelTable[][2] = [ _
    ["{F2}",    $hOK], _
    ["^s",      $hSave], _
    ["^{ENTER}",$hOK], _
    ["!{ENTER}",$hApply], _
    ["^i",      $hInsert], _
    ["!t",      $hTplSave], _
    ["^m",      $hMultiEditHK], _
    ["^n",      GUICtrlCreateDummy()], _
    ["^p",      GUICtrlCreateDummy()], _
    ["^t",      GUICtrlCreateDummy()], _
    ["^a",      GUICtrlCreateDummy()], _
    ["^{BS}",   GUICtrlCreateDummy()], _
    ["{F1}",    GUICtrlCreateDummy()], _
    ["{F5}",    GUICtrlCreateDummy()], _
    ["{F6}",    GUICtrlCreateDummy()], _
    ["{F7}",    GUICtrlCreateDummy()], _
    ["!{UP}",   GUICtrlCreateDummy()], _
    ["!{DOWN}", GUICtrlCreateDummy()], _
    ["!{HOME}", GUICtrlCreateDummy()], _
    ["{F12}",   GUICtrlCreateDummy()], _
    ["^{F12}",  GUICtrlCreateDummy()]]

    ; index of first hotkey not assigned to existing control (button etc)
    Local $nMinHKIndex = 7

    ; assign template accelerators
    Local $nTplStartHKIndex = UBound($hGUI_AccelTable)
    ReDim $hGUI_AccelTable[ $nTplStartHKIndex + 20 ][2]
    Local $nMaxHKIndex = UBound($hGUI_AccelTable)-1
    For $iKey = 0 To 19
        $hGUI_AccelTable[$nTplStartHKIndex + $iKey][0] = "^" & ($iKey > 9 ? "+" & ($iKey - 10) : $iKey)
        $hGUI_AccelTable[$nTplStartHKIndex + $iKey][1] = GUICtrlCreateDummy()
    Next

    ; notifications
    Local $hEnChangeNotify = GUICtrlCreateDummy()
    Local $hInsertNotify   = GUICtrlCreateDummy()
    Local $hDropdownNotify = GUICtrlCreateDummy()
    Local $hModifyNotify   = GUICtrlCreateDummy()

    ; set default encoding
    _SetEncCombo($hEncSwitch, $g_bNeedUnicode)

    ; set default description
    Local $sDefDescr = _GetDefDescr($aDescr, $g_aList)
    GUICtrlSetData($hEdit, $sDefDescr)
    GUICtrlSetData($hSize, StringLen($sDefDescr))

    ; apply default description with empty filename, it will be common description for multiple files
    If $g_sCurrFile = "" Then _DescrIon_SetDescr($aDescr, $g_sCurrFile, $sDefDescr)

    ; default insert/replace mode
    Local $nDefaultMode = IniRead($g_sScriptConfig, "Main", "DefaultMode", 0)
    If $nDefaultMode < 0 Or $nDefaultMode > 2 Then $nDefaultMode = 0
    GUICtrlSetState($ahOptions[$nDefaultMode], $GUI_CHECKED)

    _SetFocus($hEditWnd)

    ; should be called after focus is set
    _GUISetAccelerators($hGUI_AccelTable)

    _TCHSetGUIDefaultIcon()

    _GUI_CenterToParent($hParent, $hGUI)

    Local $hIconSingle = Resource_Icon_Single_16()
    Local $hIconMulti  = Resource_Icon_Multi_16()

    GUICtrlSetTip($hMultiEdit, $LANG("31"))
    _WindowSetImageByHandle($hMultiEdit, $hIconMulti)

    Local $tFlag = DllStructCreate("boolean Block") ; initialized as False

    _MSGHandler_Register($hGUI, $WM_COMMAND, _WM_COMMAND_FUNC, _MSGHandler_MakeData($hEdit, $hEnChangeNotify))
    _MSGHandler_Register($hGUI, $WM_NOTIFY,  _WM_NOTIFY_FUNC,  _MSGHandler_MakeData($hOK, $hDropdownNotify, DllStructGetPtr($tFlag)))
    _MSGHandler_Register($hGUI, $WM_SYSCOMMAND, _GetHelp)

    GUISetState(@SW_SHOW)

    _GUI_AdjustSize($hGUI)

    ; disable total commander window, cause a problem if the script is crashed :)
    _TCHSetParentModal($hParent)

    Local $nMsg
    Local $bMultiEdit, $bState
    Local $sText

    While 1
        $nMsg = GUIGetMsg()
        Switch $nMsg
            Case $GUI_EVENT_CLOSE, $hCancel
                ExitLoop

            Case $hEditComm
                If $bMultiFile Then
                    If _IsComboMultiEdit($hEditComm) Then
                        $g_sCurrFile = ""
                        GUICtrlSetTip($hEditComm, "")
                        GUICtrlSetState($hEdit, _IsChecked($hMultiEdit) ? $GUI_ENABLE : $GUI_DISABLE)
                    Else
                        $g_sCurrFile = GUICtrlRead($hEditComm)
                        GUICtrlSetTip($hEditComm, $g_sCurrFile)
                        GUICtrlSetState($hEdit, $GUI_ENABLE)
                    EndIf
                    Local $sCurDescr = _DescrIon_GetDescr($aDescr, $g_sCurrFile)
                    GUICtrlSetData($hEdit, $sCurDescr)
                    GUICtrlSetData($hSize, StringLen($sCurDescr))
                    _SetFocus($hEditWnd) ; GUICtrlGetHandle($hEdit)

                    ; possibly it should be checked in EN_CHANGE?
                    $bState = _UnicodeState($g_sCurrFile, $sCurDescr)
                    _UpdateEncState($bState)
                    _SetEncComboState($hEncSwitch, $bState)
                EndIf

            Case $hMultiEditHK
                GUICtrlSetState($hMultiEdit, _IsChecked($hMultiEdit) ? $GUI_UNCHECKED : $GUI_CHECKED)
                ContinueCase

            Case $hMultiEdit
                Local $bMultiEdit = _IsChecked($hMultiEdit)

                GUICtrlSetTip($hMultiEdit, $bMultiEdit ? $LANG("31") : $LANG("32"))
                _WindowSetImageByHandle($hMultiEdit, $bMultiEdit ? $hIconMulti : $hIconSingle)

                _SetFocus($hEditWnd)

                If Not $bMultiEdit And $bMultiFile And _IsComboMultiEdit($hEditComm) Then
                    GUICtrlSetState($hEdit, $GUI_DISABLE)
                Else
                    GUICtrlSetState($hEdit, $GUI_ENABLE)
                EndIf

                _SetFocus($hEditWnd)

            Case $hEncSwitch
                $g_nNewEnc = _GetEncCombo($hEncSwitch, _UnicodeState())
                $g_bFallbackANSI = $g_nNewEnc = $ENC_ANSI
                GUICtrlSendToDummy($hModifyNotify, 1)

            Case $hOK, $hApply, $hSave
                ; if no comments modified Strl+S should not fire
                If $nMsg = $hSave And Not $bModified Then ContinueLoop

                ; ==============================================================================================
                ; Apply
                ; ==============================================================================================

                Local $sNewSrcDescr = GUICtrlRead($hEdit)
                Local $sCurDescr, $sNewDescr
                Local $bExpandMacro = _IsChecked($hMacro)

                ; expand macro in all individually modified descriptions in multiselection or for single file mode
#cs
                If $bExpandMacro Then
                    For $i = 1 To $aDescr[0][0]
                        ; only modified, also skip special multi-file description
                        If $aDescr[$i][$DI_MODIFIED] And $aDescr[$i][$DI_FILENAME] <> "" Then
                            $aDescr[$i][$DI_DESCRIPTION] = _ExpandMacro($aDescr[$i][$DI_DESCRIPTION], $g_sCurrPath & $aDescr[$i][$DI_FILENAME])
                        EndIf
                    Next
                EndIf
#ce

                ; adding/replacing comments in multi-mode
                If _IsChecked($hMultiEdit) And $bMultiFile And _IsComboMultiEdit($hEditComm) Then
                    For $i = 0 To UBound($g_aList)-1
                        ; if we by mistake selected descript.ion itself (Allow?)
                        If $g_aList[$i] = $g_sDescrIon Then ContinueLoop

                        ; expand macro for selected files
                        $sNewDescr = $bExpandMacro ? _ExpandMacro($sNewSrcDescr, $g_sCurrPath & $g_aList[$i]) : $sNewSrcDescr

                        ; do insert/replace
                        If _IsChecked($ahOptions[0]) Then ; replace
                            _DescrIon_SetDescr($aDescr, $g_aList[$i], $sNewDescr)
                        Else
                            $sCurDescr = _DescrIon_GetDescr($aDescr, $g_aList[$i])
                            If _IsChecked($ahOptions[1]) Then ; insert to end
                                _DescrIon_SetDescr($aDescr, $g_aList[$i], $sCurDescr & $sNewDescr)
                            Else
                                _DescrIon_SetDescr($aDescr, $g_aList[$i], $sNewDescr & $sCurDescr)
                            EndIf
                        EndIf
                    Next

                    ; switch mode to single
                    GUICtrlSetState($hMultiEdit, $GUI_UNCHECKED)
                    GUICtrlSetState($hEdit, $GUI_DISABLE)
                    GUICtrlSetTip($hMultiEdit, $LANG("32"))
                    _WindowSetImageByHandle($hMultiEdit, $hIconSingle)

                    $sDefDescr = _GetDefDescr($aDescr, $g_aList)
                    _DescrIon_SetDescr($aDescr, $g_sCurrFile, $sDefDescr)
                Else
                    ; apply current single file
                    _DescrIon_SetDescr($aDescr, $g_sCurrFile, $bExpandMacro ? _ExpandMacro($sNewSrcDescr, $g_sCurrPath & $g_sCurrFile) : $sNewSrcDescr)
                    $sDefDescr = _DescrIon_GetDescr($aDescr, $g_sCurrFile)
                EndIf
                ; set new default description
                GUICtrlSetData($hEdit, $sDefDescr)

                GUICtrlSendToDummy($hModifyNotify, 1)

                If $nMsg = $hApply Then ContinueLoop
                
                ; ==============================================================================================
                ; Save
                ; ==============================================================================================

                $g_nNewEnc = _GetEncCombo($hEncSwitch, _UnicodeState())
                _DescrIon_SetMeta($aDescr, $DI_META_ENCODING, $g_nNewEnc)

                ; if we upgrade from the ANSI to Unicode, and there was broken names, try to find and replace them
                If _DescrIon_NeedNameCheck($aDescr) And ($g_nNewEnc <> $ENC_ANSI) Then
                    For $i = 0 To UBound($g_aList)-1
                        _DescrIon_ReplaceAnsiName($aDescr, $g_aList[$i])
                    Next
                EndIf

                _DescrIon_Save($g_sDescrIon, $aDescr)
                _SetFocus($hEditWnd) ; GUICtrlGetHandle($hEdit)
                _TCHSendMsg(540)

                $g_nCurEnc = $g_nNewEnc
                GUICtrlSetData($hPrevEnc, _GetEncString($g_nNewEnc))

                GUICtrlSendToDummy($hModifyNotify, 0)

                If $nMsg = $hSave Then ContinueLoop

                ; ==============================================================================================
                ; Close window
                ; ==============================================================================================
                
                ExitLoop

            Case $hEnChangeNotify
                $sText = GUICtrlRead($hEdit)
                $bState = _UnicodeState($g_sCurrFile, $sText)
                GUICtrlSetData($hSize, StringLen($sText))

                If $bMultiFile And Not _IsComboMultiEdit($hEditComm) Then
                    GUICtrlSetState($hMultiEdit, $GUI_UNCHECKED)
                    GUICtrlSetTip($hMultiEdit, $LANG("31"))
                    _WindowSetImageByHandle($hMultiEdit, $hIconSingle)
                EndIf

                If $g_sCurrFile <> "" Then GUICtrlSendToDummy($hModifyNotify, 1)

                _UpdateEncState($bState)
                _SetEncComboState($hEncSwitch, $bState)
                _DescrIon_SetDescr($aDescr, $g_sCurrFile, $sText)

            Case $hModifyNotify
                $bState = GUICtrlRead($hModifyNotify)
                If $bState = $bModified Then ContinueLoop
                $bModified = $bState
                GUICtrlSetState($hSave, $bModified ? $GUI_ENABLE : $GUI_DISABLE)
                WinSetTitle($hGUI, "", $bModified ? $LANG("1") & " [*]" : $LANG("1"))

            Case $hDropdownNotify
                Local $aPos = ControlGetPos($hGUI, "", $hOK)
                GUICtrlSetState($hTplSave, GUICtrlRead($hSize) = "0" ? $GUI_DISABLE : $GUI_ENABLE)
                _ClientToScreen($hGUI, $aPos[0], $aPos[1])
                _TrackPopupMenu($hGUI, GUICtrlGetHandle($hOKMenu), $aPos[0] + $aPos[2], $aPos[1] + $aPos[3], $TPM_RIGHTALIGN)
                $tFlag.Block = False

            Case $hOptions
                Local $nOptState = ($g_sCurrFile = "" And _IsChecked($hMultiEdit)) ? $GUI_ENABLE : $GUI_DISABLE
                GUICtrlSetState($ahOptions[0], $nOptState)
                GUICtrlSetState($ahOptions[1], $nOptState)
                GUICtrlSetState($ahOptions[2], $nOptState)

                Local $aPos = ControlGetPos($hGUI, "", $hOptions)
                _ClientToScreen($hGUI, $aPos[0], $aPos[1])
                _TrackPopupMenu($hGUI, GUICtrlGetHandle($hOptMenu), $aPos[0], $aPos[1]+$aPos[3])

            Case $hMacro
                GUICtrlSetState($hMacro, _IsChecked($hMacro) ? $GUI_UNCHECKED : $GUI_CHECKED)

            Case $hInsertNotify
                Local $nCode = GUICtrlRead($hInsertNotify)
                Local $bInsert
                If $nCode > 2 Then
                    $nIndex = BitAND($nCode, 0xFF)
                    $sText = IniRead($g_sScriptConfig, "Templates", $nIndex, "<error>")
                    If $sText = "<error>" Then ContinueLoop
                    $sText = _TransformText($sText, 1)
                    $bInsert = False
                Else
                    Local $sText = _InsertDialog($hGUI, GUICtrlRead($hInsertNotify))
                    If @extended = -1 Then
                        _SetFocus($hEditWnd)
                        ContinueLoop
                    EndIf
                    $bInsert = @extended = 1 ; only for plugins
                EndIf

                If $sText <> "" Then
                    If $bInsert Then
                        GUICtrlSendMsg($hEdit, $EM_REPLACESEL, 1, $sText)
                        $sText = GUICtrlRead($hEdit)
                        Local $nLen = StringLen($sText)
                        GUICtrlSetData($hSize, $nLen)
                    Else
                        GUICtrlSetData($hEdit, $sText)
                        Local $nLen = StringLen($sText)
                        GUICtrlSetData($hSize, $nLen)
                    EndIf
                EndIf

                _SetFocus($hEditWnd)
                _DescrIon_SetDescr($aDescr, $g_sCurrFile, $sText)
                $bModified = True

            Case $hInsert
                GUICtrlSendToDummy($hInsertNotify, -1)

            Case $hTplSave
                $sText = GUICtrlRead($hEdit)
                If $sText = "" Then ContinueLoop
                Local $nLastTemplate = 0 ; first template
                Local $aTemplates = IniReadSection($g_sScriptConfig, "Templates")
                If Not @error Then $nLastTemplate = $aTemplates[ $aTemplates[0][0] ][0]
                IniWrite($g_sScriptConfig, "Templates", $nLastTemplate + 1, _TransformText($sText, 8))
                _GUI_NotifyWindow($hGUI, $LANG("28"), 250, 45, 1500)


            ; Hotkey handler
            Case $hGUI_AccelTable[$nMinHKIndex][1] To $hGUI_AccelTable[$nMaxHKIndex][1]
                For $i = $nMinHKIndex To $nMaxHKIndex
                    If $nMsg = $hGUI_AccelTable[$i][1] Then
                        Switch $i
                            Case $nTplStartHKIndex To $nMaxHKIndex
                                Local $iKey = $i - $nTplStartHKIndex
                                Local $nIndex = $iKey = 0 ? 10 : $iKey = 10 ? 20 : $iKey
                                GUICtrlSendToDummy($hInsertNotify, BitOR($nIndex, 0xFF00))

                            Case Else
                                ; use strings instead indexes here, else it's a pain to add hotkeys
                                Switch $hGUI_AccelTable[$i][0]
                                    Case "^n", "{F5}"
                                        GUICtrlSendToDummy($hInsertNotify, 0)

                                    Case "^p", "{F6}"
                                        GUICtrlSendToDummy($hInsertNotify, 1)

                                    Case "^t", "{F7}"
                                        GUICtrlSendToDummy($hInsertNotify, 2)

                                    Case "^a"
                                        GUICtrlSendMsg($hEdit, $EM_SETSEL, 0, -1)

                                    Case "!{UP}" ; Previous file
                                        Local $nIndex = GUICtrlSendMsg($hEditComm, $CB_GETCURSEL, 0, 0)
                                        If $nIndex <> 0 Then
                                            $nIndex -= 1
                                            GUICtrlSendMsg($hEditComm, $CB_SETCURSEL, $nIndex, 0)
                                            _TCHSendMessage($hGUI, $WM_COMMAND, BitOR($hEditComm, 0x00010000), GUICtrlGetHandle($hEditComm))
                                        EndIf

                                    Case "!{DOWN}" ; Next file
                                        Local $nIndex = GUICtrlSendMsg($hEditComm, $CB_GETCURSEL, 0, 0)
                                        Local $nCount = GUICtrlSendMsg($hEditComm, $CB_GETCOUNT, 0, 0)
                                        If $nIndex <> $nCount Then
                                            $nIndex += 1
                                            GUICtrlSendMsg($hEditComm, $CB_SETCURSEL, $nIndex, 0)
                                            _TCHSendMessage($hGUI, $WM_COMMAND, BitOR($hEditComm, 0x00010000), GUICtrlGetHandle($hEditComm))
                                        EndIf

                                    Case "!{HOME}" ; Multiedit
                                        GUICtrlSendMsg($hEditComm, $CB_SETCURSEL, 0, 0)
                                        _TCHSendMessage($hGUI, $WM_COMMAND, BitOR($hEditComm, 0x00010000), GUICtrlGetHandle($hEditComm))

                                    Case "{F1}"
                                        _HelpDialog($hGUI, _HelpText())

                                    Case "^{BS}"
                                        ; not quite correct, hotkey should be designated to edit
                                        If _GetFocus() = $hEditWnd Then ; GUICtrlGetHandle($hEdit)
                                            Local $sText = GUICtrlRead($hEdit)
;                                            Local $sNewText = StringRegExpReplace($sText, "(?s)^(.*?\s*?)\S+?\s*$", "\1")
                                            GUICtrlSendMsg($hEdit, $EM_SETSEL, StringLen(StringRegExpReplace($sText, "(?s)^(.*?\s*?)\S+?\s*$", "\1")), StringLen($sText))
                                            GUICtrlSendMsg($hEdit, $WM_KEYDOWN, 0x2E, 0)
;                                            GUICtrlSetData($hEdit, StringRegExpReplace($sText, "(?s)^(.*?\s*?)\S+?\s*$", "\1"))
                                        EndIf

                                    Case "{F12}" ; delete comment (clear window)
                                        If _GetFocus() = $hEditWnd Then 
                                            GUICtrlSendMsg($hEdit, $EM_SETSEL, 0, -1)
                                            ; EM_REPLACESEL is not enable undo for this operation, 
                                            ; so delete selection by sending VK_DELETE to contol
                                            GUICtrlSendMsg($hEdit, $WM_KEYDOWN, 0x2E, 0)
                                        EndIf

                                    Case "^{F12}" ; delete comment (clear window) and close window
                                        If _GetFocus() = $hEditWnd Then
                                            GUICtrlSetData($hEdit, "")
                                            _TCHSendMessage($hGUI, $WM_COMMAND, BitOR($hOK, 0x00010000), 0)
                                        EndIf
                                EndSwitch
                        EndSwitch
                    EndIf
                Next
        EndSwitch
    WEnd
EndFunc

Func _IsChecked($hCtrlID)
    Return BitAND(GUICtrlRead($hCtrlID), $GUI_CHECKED) = $GUI_CHECKED
EndFunc

Func _GetComboIndex($hCtrlID)
    Return GUICtrlSendMsg($hCtrlID, $CB_GETCURSEL, 0, 0)
EndFunc

Func _IsComboMultiEdit($hCtrlID)
    Return GUICtrlSendMsg($hCtrlID, $CB_GETCURSEL, 0, 0) = 0
EndFunc

; ==================================================================================================
; Insert dialog functions
; ==================================================================================================
Func _InsertDialog($hParent, $nTabIndex)

    _InitPluginsArray($aPlugins)

    Local Static $sPlugCurName = ""
    Local Static $nLocalTabIndex = IniRead($g_sScriptConfig, "Main", "InsertTabIndex", -1)
    Local Static $bSaveTabIndex = $nLocalTabIndex <> -1
    If $nLocalTabIndex = -1 Then $nLocalTabIndex = 0

    Local $bModified = False

    ; ==============================================================================================
    ; client area on tabs, unfinished
    Local $nMargin = 5, $nCtrlHeight = 25, $nCtrlWidth = 75
    Local $nLeft = $nMargin * 2 + 1
    Local $nTop  = $nMargin + 26
    Local $nWidth  = 580
    Local $nHeight = 280

    Local $hInsertGUI = GUICreate($LANG("20"), $nWidth + $nMargin * 5, $nTop + $nHeight + $nMargin * 3 + $nCtrlHeight, -1, -1, BitOR($WS_POPUP, $WS_CAPTION, $WS_SYSMENU), -1, $hParent)
    Local $hTab = GUICtrlCreateTab($nMargin, $nMargin, $nLeft + $nWidth + $nMargin, $nTop + $nHeight)
    Local $ahTabs[3]

    ; ==============================================================================================
    $ahTabs[0] = GUICtrlCreateTabItem($LANG("23")) ; Neighbours

    Local $hList = GUICtrlCreateListView("Description", $nLeft, $nTop, $nWidth, $nHeight - 3, BitOR($GUI_SS_DEFAULT_LISTVIEW, $LVS_NOCOLUMNHEADER, $LVS_SHOWSELALWAYS, $WS_TABSTOP), BitOR($WS_EX_CLIENTEDGE,$LVS_EX_FULLROWSELECT, $LVS_EX_INFOTIP))

    Local $hAfterControl = GUICtrlCreateLabel("", 0, 0)
    DllCall("user32.dll", "bool", "SetWindowPos", "hwnd", GUICtrlGetHandle($hList), "hwnd", GUICtrlGetHandle($hAfterControl), _
                                                  "int", 0, "int", 0, "int", 0, "int", 0, "uint", BitOR(0x1, 0x2, 0x40)) ; SWP_NOSIZE + SWP_NOMOVE + SWP_SHOWWINDOW
    GUICtrlDelete($hAfterControl)

    Local $aUniqueDescr[ $aDescr[0][0] ], $nCount = 0
    For $i = 1 To $aDescr[0][0]
        ; skip "default" item with empty name or just empty descriptions
        If $aDescr[$i][0] = "" Or $aDescr[$i][1] = "" Then ContinueLoop
        ; skip file that being editing
        If $aDescr[$i][0] = $g_sCurrFile Then ContinueLoop
        For $j = 0 To $nCount - 1
            If $aUniqueDescr[$j] = $aDescr[$i][1] Then ContinueLoop(2)
        Next
        $nCount += 1
        $aUniqueDescr[$nCount-1] = $aDescr[$i][1]
    Next
    ReDim $aUniqueDescr[ $nCount ]

    _ArrayQuickSort1D($aUniqueDescr, 0, UBound($aUniqueDescr)-1)

    Local $ahNeighbour[ $nCount ]
    For $i = 0 To $nCount - 1
        $ahNeighbour[$i] = GUICtrlCreateListViewItem(_TransformText($aUniqueDescr[$i], 4), $hList) 
    Next
    GUICtrlSendMsg($hList, $LVM_SETCOLUMNWIDTH, 0, $nWidth - DllCall("user32.dll", "dword", "GetSystemMetrics", "int", 2)[0] - 5)

    ; the only way to make listview to respect tabstop...
    If UBound($ahNeighbour) > 0 Then GUICtrlSetState($ahNeighbour[0], $GUI_FOCUS)
    __GUICtrlListView_SetItemSelected(GUICtrlGetHandle($hList), 0, True)

    ; ==============================================================================================
    $ahTabs[1] = GUICtrlCreateTabItem($LANG("21")) ; Plugins

    Local $hPlugins = GUICtrlCreateCombo("", $nLeft, $nTop, 381, $nCtrlHeight, BitOR($CBS_DROPDOWNLIST,$CBS_AUTOHSCROLL,$WS_VSCROLL))
    Local $hColumns = GUICtrlCreateList("", $nLeft, 57, 249, $nHeight - $nCtrlHeight - 1, BitOR($LBS_NOTIFY,$WS_VSCROLL))
    Local $hUnits = GUICtrlCreateList("", $nLeft + 252, 57, 129, $nHeight - $nCtrlHeight - 1, BitOR($LBS_NOTIFY,$WS_VSCROLL))

    GUICtrlCreateGroup($LANG("29"), $nLeft + 385, $nTop, 195, $nHeight * 0.49)
    Local $hPreview = GUICtrlCreateLabel("", $nLeft + 390, $nTop+16, 180, $nHeight * 0.4, $SS_EDITCONTROL)
    
    GUICtrlCreateGroup($LANG("30"), $nLeft + 385, $nTop + $nHeight * 0.5, 195, $nHeight * 0.49)
    Local $hPreviewValue = GUICtrlCreateLabel("", $nLeft + 390, $nTop+16 + $nHeight * 0.5, 180, $nHeight * 0.4, $SS_EDITCONTROL)

    Local $iPlugIndex = 0

    For $i = 1 To $aPlugins[0][0]
        If $sPlugCurName = $aPlugins[$i][0] Then $iPlugIndex = $i-1
        GUICtrlSendMsg($hPlugins, $CB_ADDSTRING, -1, $aPlugins[$i][0])
    Next

    GUICtrlSendMsg($hPlugins, $CB_SETCURSEL, $iPlugIndex, 0)

    If $sPlugCurName = "" Then $sPlugCurName = $aPlugins[1][0]

    ; ==============================================================================================
    $ahTabs[2] = GUICtrlCreateTabItem($LANG("22")) ; Templates

    Local $hTplList = GUICtrlCreateListView("Template" & $cSep & "Hotkey" & $cSep & "ID", $nLeft, $nTop, $nWidth - $nCtrlHeight - $nMargin * 2, $nHeight-3, _
                                            BitOR($GUI_SS_DEFAULT_LISTVIEW, $LVS_NOCOLUMNHEADER, $LVS_SHOWSELALWAYS, $WS_TABSTOP), BitOR($WS_EX_CLIENTEDGE,$LVS_EX_FULLROWSELECT, $LVS_EX_INFOTIP))

    ; Looks like Autoit has a bug: while creating ListView on tab page, it's actually created *before* Tab control
    ; so it cause a problem with switching focus by Tab key.
    ; Solution: create "supporting" hidden control (actually it can be control on previous tab page, but it less safe) 
    ; and move ListView window after it.
    Local $hAfterControl = GUICtrlCreateLabel("", 0, 0)
    DllCall("user32.dll", "bool", "SetWindowPos", "hwnd", GUICtrlGetHandle($hTplList), "hwnd", GUICtrlGetHandle($hAfterControl), _
                                                  "int", 0, "int", 0, "int", 0, "int", 0, "uint", BitOR(0x1, 0x2, 0x40)) ; SWP_NOSIZE + SWP_NOMOVE + SWP_SHOWWINDOW
    GUICtrlDelete($hAfterControl)
    ; End of fix

    Local $nBtnLeft = $nWidth - $nCtrlHeight + $nMargin*2
    Local $hDelete = GUICtrlCreateButton(ChrW(0x2715), $nBtnLeft, $nTop + $nCtrlHeight * 2, $nCtrlHeight, $nCtrlHeight)
    GUICtrlSetFont(-1, 14, 800, 0, "Arial")
    GUICtrlSetTip(-1, $LANG("24"))
    Local $hUp   = GUICtrlCreateButton(Chr(225), $nBtnLeft, $nTop + $nCtrlHeight * 4, $nCtrlHeight, $nCtrlHeight)
    GUICtrlSetFont(-1, 10, 800, 0, "Wingdings")
    Local $hDown = GUICtrlCreateButton(Chr(226), $nBtnLeft, $nTop + $nCtrlHeight * 5 + $nMargin, $nCtrlHeight, $nCtrlHeight)
    GUICtrlSetFont(-1, 10, 800, 0, "Wingdings")

    Local $sItem
    Local $ahTemplates[0] ; empty
    Local $aTemplates = IniReadSection($g_sScriptConfig, "Templates")
    If Not @error Then
        ReDim $ahTemplates[ $aTemplates[0][0] ]

        For $i = 1 To $aTemplates[0][0]
            ; keep original array index in 3-rd hidden column,
            ; this will allow to retrieve correct template text after reorder and deleting items
            $sItem = _TransformText($aTemplates[$i][1], 1+4) & $cSep & $cSep & $i
            $ahTemplates[$i-1] = GUICtrlCreateListViewItem($sItem, $hTplList)
        Next

        _UpdateHotkeys($ahTemplates)
    EndIf

    GUICtrlSendMsg($hTplList, $LVM_SETCOLUMNWIDTH, 0, 440)
    GUICtrlSendMsg($hTplList, $LVM_SETCOLUMNWIDTH, 1, -2)
    GUICtrlSendMsg($hTplList, $LVM_SETCOLUMNWIDTH, 2, 0)

    ; the only way to make listview to respect tabstop...
    If UBound($ahTemplates) > 0 Then GUICtrlSetState($ahTemplates[0], $GUI_FOCUS)
    ; Select, but de-focus item
    __GUICtrlListView_SetItemSelected(GUICtrlGetHandle($hTplList), 0, True)

    GUICtrlCreateTabItem("")
    ; ==============================================================================================

    Local $hOK     = GUICtrlCreateButton("OK", $nLeft + $nWidth * 0.5 - $nCtrlWidth, $nTop + $nHeight + $nMargin * 2, $nCtrlWidth, $nCtrlHeight, $BS_DEFPUSHBUTTON)
    Local $hCancel = GUICtrlCreateButton($LANG("7"), $nWidth * 0.5 + $nMargin * 3, $nTop + $nHeight + $nMargin * 2, $nCtrlWidth, $nCtrlHeight, 0)

    ; should NOT be on tabs
    Local $hPluginSelMsg = GUICtrlCreateDummy()
    Local $hTabHotkey    = GUICtrlCreateDummy()
    Local $hArrowDown    = GUICtrlCreateDummy()

    Local $hGUI_AccelTable[][2] = [["{ENTER}", $hOK], ["^{TAB}", $hTabHotkey], ["!{UP}", $hUp], ["!{DOWN}", $hDown], ["{DEL}", $hDelete]]
    GUISetAccelerators($hGUI_AccelTable)

    _TCHSetGUIDefaultIcon()

    _GUI_CenterToParent($hParent, $hInsertGUI)

    _MSGHandler_RegisterDoubleClick($hInsertGUI, $hList,    $hOK)
    _MSGHandler_RegisterDoubleClick($hInsertGUI, $hColumns, $hOK)
    _MSGHandler_RegisterDoubleClick($hInsertGUI, $hUnits,   $hOK)
    _MSGHandler_RegisterDoubleClick($hInsertGUI, $hTplList, $hOK)

    GUISetState(@SW_SHOW)
    GUISetState(@SW_DISABLE, $hParent)

    _GUI_AdjustSize($hInsertGUI)

    ; ==============================================================================================

    ; init filling last plugin columns
    GUICtrlSendToDummy($hPluginSelMsg)

    ; if not called by hotkey, set locally remembered index
    If $nTabIndex = -1 Then $nTabIndex = $nLocalTabIndex
    If $nTabIndex >= 0 And $nTabIndex < UBound($ahTabs) Then
        GUICtrlSetState($ahTabs[$nTabIndex], $GUI_SHOW)
    EndIf

    Local $nMsg, $hSelectedItem, $nCurTab, $sTemp
    Local $aColumns, $aUnits, $aPlug, $sCurColumn
    Local $sResult = "", $nResType = -1

    While 1
	    $nMsg = GUIGetMsg()
        Switch $nMsg
            Case $GUI_EVENT_CLOSE, $hCancel
                ExitLoop

            Case $hTabHotkey
                $nCurTab = GUICtrlRead($hTab)
                GUICtrlSetState($ahTabs[ Int($nCurTab = 2 ? 0 : $nCurTab + 1) ], $GUI_SHOW)

            Case $hOK
                $nResType = GUICtrlRead($hTab)
                Switch $nResType
                    Case 0
                        $hSelectedItem = GUICtrlRead($hList)
                        If $hSelectedItem > 0 Then
                            For $i = 0 To UBound($ahNeighbour)-1
                                If $ahNeighbour[$i] = $hSelectedItem Then
                                    $sResult = $aUniqueDescr[$i]
                                    ExitLoop
                                EndIf
                            Next
                        EndIf
                        ExitLoop

                    Case 1
                        $sResult = _GetPluginMacro($aPlug, GUICtrlRead($hColumns), GUICtrlRead($hUnits))
                        ExitLoop

                    Case 2
                        $hSelectedItem = GUICtrlRead($hTplList)
                        If $hSelectedItem > 0 Then
                            For $i = 0 To UBound($ahTemplates)-1
                                If $ahTemplates[$i] = $hSelectedItem Then
                                    $sText = _GetTemplateByHandle($aTemplates, $hSelectedItem)
                                    If @error Then ExitLoop
                                    $sResult = _TransformText($sText, 1)
                                    ExitLoop
                                EndIf
                            Next
                        EndIf
                        ExitLoop
                EndSwitch

            Case $hPlugins, $hPluginSelMsg
                GUICtrlSendMsg($hColumns, $LB_RESETCONTENT, 0, 0)
                GUICtrlSendMsg($hUnits, $LB_RESETCONTENT, 0, 0)
                $sPlugCurName = GUICtrlRead($hPlugins)
                For $i = 1 To $aPlugins[0][0]
                    If $aPlugins[$i][0] = $sPlugCurName Then
                        If _WDXIsValidHandle($aPlugins[$i][2]) Then
                            $aPlug = $aPlugins[$i][2]
                        Else
                            $aPlug = _WDXPluginOpen($aPlugins[$i][1])
                            If @error Then ContinueLoop
                        EndIf
                        $aColumns = $aPlug[$PLUGIN_COLUMNS]
                        For $c = 0 To UBound($aColumns)-1
                            GUICtrlSendMsg($hColumns, $LB_ADDSTRING, -1, $aColumns[$c][$COLUMN_NAME])
                        Next
                        ExitLoop
                    EndIf
                Next
                GUICtrlSetData($hPreview, "")

            Case $hColumns
                If Not IsArray($aColumns) Then ContinueLoop
                GUICtrlSendMsg($hUnits, $LB_RESETCONTENT, 0, 0)
                $sCurColumn = GUICtrlRead($hColumns)
                For $c = 0 To UBound($aColumns)-1
                    If $aColumns[$c][$COLUMN_NAME] = $sCurColumn Then
                        If IsArray($aColumns[$c][$COLUMN_UNIT]) Then
                            $aUnits = $aColumns[$c][$COLUMN_UNIT]
                            For $u = 0 To UBound($aUnits)-1
                                GUICtrlSendMsg($hUnits, $LB_ADDSTRING, -1, $aUnits[$u])
                            Next
                        EndIf
                        ExitLoop
                    EndIf
                Next
                ContinueCase

            Case $hUnits
                GUICtrlSetData($hPreview, _GetPluginMacro($aPlug, GUICtrlRead($hColumns), GUICtrlRead($hUnits)))
                If $g_sCurrFile = "" Then
                    GUICtrlSetData($hPreviewValue, "<" & $LANG("10") & ">")
                Else
                    GUICtrlSetData($hPreviewValue, _GetPluginValue($aPlug, $g_sCurrPath & $g_sCurrFile, GUICtrlRead($hColumns), GUICtrlRead($hUnits)))
                EndIf

            Case $hDelete
                If GUICtrlRead($hTab) <> 2 Then ContinueLoop
                $hSelectedItem = GUICtrlRead($hTplList)
                If $hSelectedItem > 0 And MsgBox(4 + 32, "", $LANG("27"), 0, $hInsertGUI) = 6 Then
                    For $i = 0 To UBound($ahTemplates)-1
                        If $hSelectedItem = $ahTemplates[$i] Then
                            For $j = $i To UBound($ahTemplates) - 2
                                $ahTemplates[$j] = $ahTemplates[$j+1]
                            Next
                            ReDim $ahTemplates[ UBound($ahTemplates)-1 ]
                            GUICtrlDelete($hSelectedItem)
                            _UpdateHotkeys($ahTemplates)
                            __GUICtrlListView_SetItemSelected($hTplList, $i)
                            $bModified = True
                            ExitLoop
                        EndIf
                    Next
                EndIf
                _SetFocus(GUICtrlGetHandle($hTplList))

            Case $hUp
                If GUICtrlRead($hTab) <> 2 Then ContinueLoop
                $hSelectedItem = GUICtrlRead($hTplList)
                For $i = 0 To UBound($ahTemplates)-1
                    If $hSelectedItem = $ahTemplates[$i] Then
                        If $i = 0 Then ExitLoop
                        $sTemp = GUICtrlRead($ahTemplates[$i-1])
                        GUICtrlSetData($ahTemplates[$i-1], GUICtrlRead($hSelectedItem))
                        GUICtrlSetData($hSelectedItem, $sTemp)
                        _UpdateHotkeys($ahTemplates)
                        __GUICtrlListView_SetItemSelected($hTplList, $i - 1)
                        $bModified = True
                        ExitLoop
                    EndIf
                Next
                _SetFocus(GUICtrlGetHandle($hTplList))

            Case $hDown
                If GUICtrlRead($hTab) <> 2 Then ContinueLoop
                $hSelectedItem = GUICtrlRead($hTplList)
                For $i = 0 To UBound($ahTemplates)-1
                    If $hSelectedItem = $ahTemplates[$i] Then
                        If $i = UBound($ahTemplates)-1 Then ExitLoop
                        $sTemp = GUICtrlRead($ahTemplates[$i+1])
                        GUICtrlSetData($ahTemplates[$i+1], GUICtrlRead($hSelectedItem))
                        GUICtrlSetData($hSelectedItem, $sTemp)
                        _UpdateHotkeys($ahTemplates)
                        __GUICtrlListView_SetItemSelected($hTplList, $i+1)
                        $bModified = True
                        ExitLoop
                    EndIf
                Next
                _SetFocus(GUICtrlGetHandle($hTplList))

        EndSwitch
    Wend

    _MSGHandler_Unregister($hInsertGUI)

    $nLocalTabIndex = GUICtrlRead($hTab)
    If $bSaveTabIndex Then IniWrite($g_sScriptConfig, "Main", "InsertTabIndex", $nLocalTabIndex)

    ; save templates, if modified
    If $bModified Then
        IniDelete($g_sScriptConfig, "Templates")
        For $i = 0 To UBound($ahTemplates)-1
            $sText = _GetTemplateByHandle($aTemplates, $ahTemplates[$i])
            If @error Then ContinueLoop
            IniWrite($g_sScriptConfig, "Templates", $i + 1, $sText)
        Next
    EndIf

    GUISetState(@SW_ENABLE, $hParent)
    GUIDelete($hInsertGUI)
    GUISwitch($hParent)
    Return SetError(0, $nResType, $sResult)
EndFunc

Func _GetTemplateByHandle(ByRef $aTemplates, $hItem)
    Local $aItem = StringSplit(GUICtrlRead($hItem), $cSep)
    If $aItem[0] < 3 Then Return SetError(1)
    Local $nIndex = Int($aItem[3])
    If $nIndex >= 1 And $nIndex <= $aTemplates[0][0] Then Return $aTemplates[$nIndex][1]
    Return SetError(2)
EndFunc

Func _UpdateHotkeys(ByRef $aList)
    Local $nCount = 1
    For $i = 0 To UBound($aList)-1
        If $nCount > 20 Then ExitLoop
        Local $nKey = ($nCount = 10 Or $nCount = 20) ? 0 : $nCount > 10 ? $nCount - 10 : $nCount
        Local $sHotKey = $nCount > 10 ? "Ctrl+Shift+" & $nKey : "Ctrl+" & $nKey
        GUICtrlSetData($aList[$i], $cSep & $sHotKey & $cSep & $cSep)
        $nCount += 1
    Next
EndFunc

; ==================================================================================================
; Plugins support functions
; ==================================================================================================
; initialized when required
Func _InitPluginsArray(ByRef $aPlugins)
    If IsArray($aPlugins) Then Return ; already initialized

    Local $aPlugArray = _TCHIniReadConList($TC_INI, "ContentPlugins", "@")
    Local $aPlugs[ $aPlugArray[0][0] + 2 ][3] = [[$aPlugArray[0][0]+1,0,0]]

    Local $aExclude = StringSplit(IniRead($g_sScriptConfig, "Main", "PluginExclude", ""), ";")
    Local $aCustomPlugs = IniReadSection($g_sScriptConfig, "CustomPlugins")

    $aPlugs[1][0] = $g_sMCEPluginName
    $aPlugs[1][2] = _MCE_GetValue()

    Local $sPlugPath, $sPlugName, $nCount = 1
    For $i = 1 To $aPlugArray[0][0]
        $sPlugPath = _TCHExpandEnv($aPlugArray[$i][0])
        $sPlugName = StringRegExpReplace($sPlugPath, "(?m)^.+?\\([^\\]+)\.[^\.]+$", "\1")
        For $j = 1 To $aExclude[0]
            If $aExclude[$j] = $sPlugName Then ContinueLoop(2)
        Next
        $nCount += 1
        $aPlugs[$nCount][0] = $sPlugName
        $aPlugs[$nCount][1] = $sPlugPath
        $aPlugs[$nCount][2] = Null
    Next

    ; custom plugins adding
    If IsArray($aCustomPlugs) Then
        ReDim $aPlugs[$nCount + 1 + $aCustomPlugs[0][0]][3]
        For $i = 1 To $aCustomPlugs[0][0]
            $sPlugPath = _TCHExpandEnv($aCustomPlugs[$i][1])
            $sPlugName = StringRegExpReplace($sPlugPath, "(?m)^.+?\\([^\\]+)\.[^\.]+$", "\1")
            For $j = $nCount To 2 Step -1
                If $aPlugs[$j][0] = $sPlugName Then ContinueLoop(2) ; duplicate
            Next
            If FileExists($sPlugPath) Then
                $nCount += 1
                $aPlugs[$nCount][0] = $sPlugName
                $aPlugs[$nCount][1] = $sPlugPath
                $aPlugs[$nCount][2] = Null
            EndIf
        Next
    EndIf

    $aPlugs[0][0] = $nCount
    ReDim $aPlugs[$nCount + 1][3]

    _ArrayQuickSort2D($aPlugs, 1, 2, $nCount, 0, 2)

    $aPlugins = $aPlugs
EndFunc

Func _GetPluginMacro(ByRef $aPlug, $sColumn, $sUnit)
    If $sColumn = "" Then Return ""
    Local $sMacro = "[=" & $aPlug[$WDX_PLUGIN_NAME] & "." & $sColumn
    If $sUnit <> "" Then $sMacro &= "." & $sUnit
    $sMacro &= "]"
    Return $sMacro
EndFunc

Func _GetPluginValue(ByRef $aPlug, $sFileName, $sColumn, $sUnit)
    If $aPlug[$WDX_PLUGIN_NAME] = $g_sMCEPluginName Then
        Return _MCE_GetValue($sFileName, $sColumn, $sUnit)
    Else
        If $sUnit = "" Then $sUnit = 0
        Return _WDXGetValue($aPlug, $sFileName, $sColumn, $sUnit)
    EndIf
EndFunc

; Internal pseudo-plugin similar to TC's "tc" plugin
Func _MCE_GetValue($vCurrFile = Null, $vColumn = Null, $vUnit = Null)
    ; Define units
    Local Static $aDateFormat[] = ["YMD hms", "Y-M-D h-m-s", "D.M.Y h:m:s", "D-M-Y h:m:s", "M/D/Y h:m:s", "D/M/Y h:m:s", "Y/M/D h:m:s", _
                                   "YMD", "Y-M-D", "D.M.Y", "D-M-Y", "M/D/Y", "D/M/Y", "Y/M/D", "Y", "M", "D"]
    Local Static $aTimeFormat[] = ["hms", "h.m.s", "h:m:s", "h.m", "h:m", "a.m.sP", "h", "m", "s"]
    Local Static $aSizeFormat[] = ["bytes", "kbytes", "Mbytes", "Gbytes"]

    ; column types are not required here, so make them zero
    Local Static $aColumns[][$COLUMN_MAXINDEX] = [ _
    [0, "FullName",      ""], _
    [0, "Name",          ""], _
    [0, "Ext",           ""], _
    [0, "DOSName",       ""], _
    [0, "Path",          ""], _
    [0, "Size",          $aSizeFormat], _
    [0, "CreationDate",  $aDateFormat], _
    [0, "CreationTime",  $aTimeFormat], _
    [0, "WriteDate",     $aDateFormat], _
    [0, "WriteTime",     $aTimeFormat], _
    [0, "AccessDate",    $aDateFormat], _
    [0, "AccessTime",    $aTimeFormat], _
    [0, "Attributes",    ""], _
    [0, "AttributeStr",  ""], _
    [0, "Archive",       ""], _
    [0, "Read Only",     ""], _
    [0, "Hidden",        ""], _
    [0, "System",        ""], _
    [0, "Directory",     ""], _
    [0, "Compressed",    ""], _
    [0, "Encrypted",     ""], _
    [0, "Offline",       ""], _
    [0, "VersionString", ""], _
    [0, "Random Number", ""]]

    ; create a handle that will be valid for handle check function
    Local Static $aHandle[3] = [0xFF, $aColumns, $g_sMCEPluginName]

    If $vCurrFile = Null Then Return $aHandle
    If $vColumn   = Null Then Return SetError(1, 0, "")

    ; it's way better to work with integers in Switch, get string index
    Local $nIndex = -1
    For $i = 0 To UBound($aColumns) - 1
        If $aColumns[$i][1] = $vColumn Then
            $nIndex = $i
            ExitLoop
        EndIf
    Next

    Switch $nIndex
        Case 0
            Return _TCHExtractFileName($vCurrFile, False)
        Case 1
            Return _TCHExtractFileName($vCurrFile, True)
        Case 2
            Return _TCHExtractFileExt($vCurrFile)
        Case 3
            Return _TCHExtractFileName(FileGetShortName($vCurrFile), False)
        Case 4
            Return _TCHExtractFilePath($vCurrFile)
        Case 5
            Local $nSize = FileGetSize($vCurrFile)
            Switch $vUnit
                Case $aSizeFormat[1]
                    Return Round($nSize / 0x400)
                Case $aSizeFormat[2]
                    Return Round($nSize / 0x100000)
                Case $aSizeFormat[3]
                    Return Round($nSize / 0x40000000)
                Case Else
                    Return $nSize
            EndSwitch
        Case 6 To 11
            Local $aDate, $sResult = $vUnit
            Switch $nIndex
                Case 6, 7
                    $aDate = FileGetTime($vCurrFile, 0)
                Case 8, 9
                    $aDate = FileGetTime($vCurrFile, 1)
                Case 10, 11
                    $aDate = FileGetTime($vCurrFile, 2)
            EndSwitch

            If ($vUnit = Null) Or ($vUnit = "") Then
                ; seems it should be locale-dependent format?
                Switch $nIndex
                    Case 6, 8, 10
                        $sResult = "D.M.Y h:m"
                    Case 7, 9, 11
                        $sResult = "h:m:s"
                EndSwitch
            EndIf
            $sResult = StringReplace($sResult, "Y", $aDate[0], 0, 1)
            $sResult = StringReplace($sResult, "M", $aDate[1], 0, 1)
            $sResult = StringReplace($sResult, "D", $aDate[2], 0, 1)
            $sResult = StringReplace($sResult, "h", $aDate[3], 0, 1)
            $sResult = StringReplace($sResult, "m", $aDate[4], 0, 1)
            $sResult = StringReplace($sResult, "s", $aDate[5], 0, 1)
            $sResult = StringReplace($sResult, "a", ($aDate[3] > 12 ? $aDate[3] - 12 : ($aDate[3] = 0 ? 12 : $aDate[3])), 0, 1) ; :-/
            $sResult = StringReplace($sResult, "P", ($aDate[3] >= 12 ? "pm" : "am"), 0, 1)
            Return $sResult
        Case 12
            Local Static $aAttr[][] = [["R", 0x1], ["H", 0x2], ["S", 0x4], ["D", 0xF], ["A", 0x20], ["C", 0x800], ["0", 0x1000], ["X", 0x4000]]
            Local $sTemp = FileGetAttrib($vCurrFile)
            Local $nAttr = 0
            For $i = 0 To UBound($aAttr)-1
                If StringInStr($sTemp, $aAttr[$i][0]) Then $nAttr = BitOR($nAttr, $aAttr[$i][1])
            Next
            Return $nAttr
        Case 13
            Local $sAttr = "", $sTemp = FileGetAttrib($vCurrFile)
            $sAttr &= StringInStr($sTemp, "R") ? "r" : "-"
            $sAttr &= StringInStr($sTemp, "A") ? "a" : "-"
            $sAttr &= StringInStr($sTemp, "H") ? "h" : "-"
            $sAttr &= StringInStr($sTemp, "S") ? "s" : "-"
            Return $sAttr
        Case 14 To 21
            Return StringInStr(FileGetAttrib($vCurrFile), StringMid("ARHSDCXO", $nIndex - 13, 1)) ? "Yes" : "No"
        Case 22
            Return FileGetVersion($vCurrFile, "FileVersion")
        Case 23
            Return Random(1, 99999, 1) ; not sure what range TC is using
        Case Else
            Return ""
    EndSwitch
EndFunc

Func _ExpandMacro($sText, $sFileName)
    Local $aPart, $nBound, $sPlugName, $sColumn, $sUnit, $aPlug, $sValue
    Local $aPatt = StringRegExp($sText, "\[=([^\[\]]+?)\]", 3)
    If Not @error Then
        _InitPluginsArray($aPlugins)
        For $i = 0 To UBound($aPatt)-1
            $aPart = StringSplit($aPatt[$i], ".", 2)
            $nBound = UBound($aPart)
            If $nBound = 1 Then ContinueLoop ; wrong macro
            $sPlugName = $aPart[0]
            $sColumn   = $aPart[1]
            $sUnit = $nBound = 3 ? $aPart[2] : 0
            For $j = 1 To $aPlugins[0][0]
                If $aPlugins[$j][0] = $sPlugName Then
                    If _WDXIsValidHandle($aPlugins[$j][2]) Then
                        $aPlug = $aPlugins[$j][2]
                    Else
                        $aPlug = _WDXPluginOpen($aPlugins[$j][1])
                        If @error Then ContinueLoop(2)
                    EndIf
                    $sValue = _GetPluginValue($aPlug, $sFileName, $sColumn, $sUnit)
                    If @error Then ContinueLoop(2)
                    $sText = StringReplace($sText, "[=" & $aPatt[$i] & "]", $sValue)
                    ContinueLoop(2)
                EndIf
            Next
        Next
    EndIf
    Return $sText
EndFunc

; ==================================================================================================
; Encoding support functions
; ==================================================================================================
Func _UpdateEncState($bState)
;    If $bState And $g_nNewEnc = $ENC_ANSI Then $g_nNewEnc = $nDefEnc
    If $bState Then
        If $g_nNewEnc = $ENC_ANSI Then $g_nNewEnc = $nDefEnc
    Else
        If $g_bFallbackANSI Then $g_nNewEnc = $ENC_ANSI
    EndIf
EndFunc

Func _GetEncCombo($hEncSwitch, $bUnicode)
    Return GUICtrlSendMsg($hEncSwitch, $CB_GETCURSEL, 0, 0) + ($bUnicode ? 2 : 1)
EndFunc

Func _SetEncCombo($hEncSwitch, $bUnicode)
    GUICtrlSendMsg($hEncSwitch, $CB_SETCURSEL , $g_nNewEnc - ($bUnicode ? 2 : 1), 0)
EndFunc

Func _SetEncComboState($hEncSwitch, $bUnicode)
    Local $nCount = GUICtrlSendMsg($hEncSwitch, $CB_GETCOUNT, 0, 0)
    If $bUnicode Then
        If $nCount = 3 Then Return
    Else
        If $nCount = 4 Then Return
    EndIf
    GUICtrlSetData($hEncSwitch, $cSep & _GetEncComboString($bUnicode)) ; "|"
    _SetEncCombo($hEncSwitch, $bUnicode)
EndFunc

Func _GetEncString($nEnc)
    Local Static $aEnc[] = [$LANG("19"),"ANSI","UTF-16","UTF-16BE","UTF-8"]
    Return $aEnc[$nEnc]
EndFunc

Func _GetEncComboString($bUnicode)
    Local $sUnicode = "UTF-16" & $cSep & "UTF-16BE" & $cSep & "UTF-8"
    Return $bUnicode ? $sUnicode : "ANSI" & $cSep & $sUnicode
EndFunc

; would be a lot simpler and faster with Maps
Func _UnicodeState($sFile = Default, $sText = Default)
    Local Static $aList[1] = [0]
    ; global state is "Need" - always True
    If $g_bNeedUnicode Then Return True

    ; if multifile mode - check only it's text (INCORRECT)
;    If $sFile = "" Then Return _StringHasUniChars($sText)

    ; handle the single file mode
    If $sFile = Default And $aList[0] = 0 Then Return False

    For $i = 1 To $aList[0]
        If $aList[$i] = $sFile Then
            ; retrieve mode
            If $sFile = Default Then Return True
            ; adding mode
            ExitLoop
        EndIf
    Next

    Local $bHasChars = _StringHasUniChars($sText)
    If $i > $aList[0] Then
        ; file not found
        If Not $bHasChars Then Return $aList[0] <> 0
        $aList[0] += 1
        ReDim $aList[$aList[0]+1]
        $aList[$aList[0]] = $sFile
        Return True
    Else
       If $bHasChars Then Return True
       If $i < $aList[0] Then $aList[$i] = $aList[$aList[0]]
       ReDim $aList[$aList[0]]
       $aList[0] -= 1
       Return $aList[0] <> 0
    EndIf
EndFunc

; ==================================================================================================
; Misc support functions
; ==================================================================================================
; founds if description for all selected files are the same and return it
Func _GetDefDescr(ByRef $aDescr, ByRef $aList)
    Local $sComm = _DescrIon_GetDescr($aDescr, $aList[0])
    ; if first comment is empty, it's not matter if any other will have text or not
    If $sComm = "" Then Return ""
    For $i = 1 To UBound($aList) - 1
        If Not ($sComm = _DescrIon_GetDescr($aDescr, $aList[$i])) Then Return ""
    Next
    Return $sComm
EndFunc

Func _OnExit()
    ; cleanup opened plugins
    If IsArray($aPlugins) Then
        For $i = 2 To $aPlugins[0][0]
             _WDXPluginClose($aPlugins[$i][2])
        Next
    EndIf
    ; important to enable parent
    _TCHSetParentModal($TC_HWND, False)
EndFunc

Func _TransformText($sText, $nMode)
    Local $cTempChar = Chr(0x1)

    Local $sNewText = $sText

    If BitAND($nMode, 1) Then ; expand /n
        $sNewText = StringReplace(StringFormat($sNewText), @LF, @CRLF)
    EndIf

    If BitAND($nMode, 2) Then ; expand macro
        $sNewText = _ExpandMacro($sNewText, $g_sCurrFile)
    EndIf

    If BitAND($nMode, 4) Then ; prepare for lists
        $sNewText = StringReplace($sNewText, @CRLF, "  ")
        $sNewText = StringReplace($sNewText, @LF, "  ") ; unlikely
    EndIf

    If BitAND($nMode, 8) Then ; add /n back
        $sNewText = StringReplace($sNewText, "\n", $cTempChar)
        $nReplaced = @extended
        $sNewText = StringReplace($sNewText, @CRLF, "\n")
        $sNewText = StringReplace($sNewText, @LF, "\n") ; unlikely
        If $nReplaced > 0 Then $sNewText = StringReplace($sNewText, $cTempChar, "\\n")
    EndIf
    Return $sNewText
EndFunc

; ==================================================================================================
; Help dialog support functions
; ==================================================================================================
Func _HelpText()
    Local Static $T = ""
    If $T <> "" Then Return $T
    _AddHelp($T, "Hotkeys", 2, 1000)
    _AddHelp($T, "Main window:", 1, 1001)
    _AddHelp($T, "F2, Ctrl+Enter - ")
    _AddHelp($T, "apply, save and close window.", 1, 1002)
    _AddHelp($T, "Alt+Enter - ")
    _AddHelp($T, "apply changes to selected files (expanding macro).", 1, 1022)
    _AddHelp($T, "Ctrl+S - ")
    _AddHelp($T, "apply and save without closing window.", 1, 1023)
    _AddHelp($T, "Ctrl+M - ")
    _AddHelp($T, "switch multiedior or single editor modes.", 1, 1024)
    _AddHelp($T, "Ctrl+N, F5 - ")
    _AddHelp($T, "open window for inserting neighbour comment.", 1, 1003)
    _AddHelp($T, "Ctrl+P, F6 - ")
    _AddHelp($T, "open window for inserting plugins macro.", 1, 1004)
    _AddHelp($T, "Ctrl+T, F7 - ")
    _AddHelp($T, "open window for inserting template saved before.", 1, 1005)
    _AddHelp($T, "Ctrl+I - ")
    _AddHelp($T, "open window with last chosen insert method (neighbour, plugins or template).", 1, 1006)
    _AddHelp($T, "Alt+T - ")
    _AddHelp($T, "save editor text as template.", 2, 1025)
    _AddHelp($T, "Ctrl+A - ")
    _AddHelp($T, "select all text in editor.", 1, 1007)
    _AddHelp($T, "Ctrl+Z - ")
    _AddHelp($T, "revert last edit operation.", 1, 1008)
    _AddHelp($T, "Ctrl+Backspace - ")
    _AddHelp($T, "remove previous word.", 1, 1026)
    _AddHelp($T, "F12 - ")
    _AddHelp($T, "deleting editor content.", 1, 1027)
    _AddHelp($T, "Ctrl+F12 - ")
    _AddHelp($T, "deleting editor content and closing window.", 2, 1028)
    _AddHelp($T, "Alt+Up - ")
    _AddHelp($T, "switch to next single file in multifile mode.", 1, 1009)
    _AddHelp($T, "Alt+Down - ")
    _AddHelp($T, "switch to previous single file in multifile mode.", 1, 1010)
    _AddHelp($T, "Alt+Home - ")
    _AddHelp($T, "switch to multifile mode.", 2, 1011)
    _AddHelp($T, "Ctrl+1 ... Ctrl+0 - ")
    _AddHelp($T, "insert template from 1 to 10 into editor.", 1, 1012)
    _AddHelp($T, "Ctrl+Shift+1 ... Ctrl+Shift+0 - ")
    _AddHelp($T, "insert template from 11 to 20 into editor.", 2, 1013)
    _AddHelp($T, "F1 - ")
    _AddHelp($T, "this help.", 2, 1014)
    _AddHelp($T, "Insert dialog:", 1, 1015)
    _AddHelp($T, "Ctrl+Tab - ")
    _AddHelp($T, "switch tabs.", 1, 1016)
    _AddHelp($T, "Enter - ")
    _AddHelp($T, "close window and insert selected item into editor", 2, 1017)
    _AddHelp($T, "On ""Templates"" tab:", 1, 1018)
    _AddHelp($T, "Delete - ")
    _AddHelp($T, "delete selected template.", 1, 1019)
    _AddHelp($T, "Alt-Up - ")
    _AddHelp($T, "move selected template up.", 1, 1020)
    _AddHelp($T, "Alt-Down - ")
    _AddHelp($T, "move selected template down.", 1, 1021)
    Return $T
EndFunc

Func _AddHelp(ByRef $sText, $sString, $nCRLFCnt = 0, $nLangID = Default)
    Local Static $bIsENG = $TC_LANG = "ENG", $sLangFile = $LANG(LangGetFile)
    Local Static $aCRLF[] = ["", @CRLF, @CRLF & @CRLF]
    If $bIsENG Or $nLangID = Default Then
        $sText &= $sString & $aCRLF[$nCRLFCnt]
    Else
        $sText &= IniRead($sLangFile, $TC_LANG, $nLangID, $sString) & $aCRLF[$nCRLFCnt]
    EndIf
EndFunc

Func _HelpDialog($hParent, $sTopic)
    Local $nMsg
    Local $nWidth = 600, $nHeight = 500, $nMargin = 5
    Local $hHelpGUI = GUICreate($LANG("16"), $nWidth, $nHeight, -1, -1, BitOR($WS_POPUP, $WS_CAPTION, $WS_SYSMENU), Default, $hParent)
    GUICtrlCreateEdit($sTopic, $nMargin, $nMargin, $nWidth - $nMargin * 2, $nHeight - $nMargin * 2, BitOR($ES_READONLY, $WS_VSCROLL))
    GUICtrlSetBkColor(-1, DllCall("user32.dll", "dword", "GetSysColor", "int", 5)[0])
    GUICtrlSetFont(-1, 9, 400, 0, "Courier New")

    _GUI_CenterToParent($hParent, $hHelpGUI)

    GUISetState()
    GUICtrlSendMsg(-1, $EM_SETSEL, 0, 0)
    GUISetState(@SW_DISABLE, $hParent)

    _GUI_AdjustSize($hHelpGUI)

    GUISwitch($hHelpGUI)
    Do
        $nMsg = GUIGetMsg()
    Until $nMsg = $GUI_EVENT_CLOSE
    GUISetState(@SW_ENABLE, $hParent)
    GUIDelete($hHelpGUI)

    GUISwitch($hParent)
EndFunc

; ==================================================================================================
; Message handlers
; ==================================================================================================

; handler for counting edit length
Func _WM_COMMAND_FUNC($hWnd, $msg, $wParam, $lParam, $vData)
    Local $nID = BitAND($wParam, 0xFFFF)
    Local $nMsg = BitShift($wParam, 16)
    If ($nID = $vData[0]) And ($nMsg = $EN_CHANGE) Then
        GUICtrlSendToDummy($vData[1])
        Return 1
    EndIf
    Return $GUI_RUNDEFMSG
EndFunc

; handler for split button
Func _WM_NOTIFY_FUNC($hWnd, $nMsg, $wParam, $lParam, $vData)
    Local $tNMBCDROPDOWN = DllStructCreate("STRUCT;hwnd hwndFrom;uint idFrom;uint code;ENDSTRUCT;long rcButton[4]", $lParam)
    Local $nCode = Int($tNMBCDROPDOWN.code, 1) ; 32-bit integer!
    If $nCode = $BCN_DROPDOWN And $tNMBCDROPDOWN.idFrom = $vData[0] Then
        Local $tFlag = DllStructCreate("boolean Block", $vData[2])
        If Not $tFlag.Block Then GUICtrlSendToDummy($vData[1])
        $tFlag.Block = True
        Return 1
    EndIf
    Return $GUI_RUNDEFMSG
EndFunc

; handle click on ? sign in title
Func _GetHelp($hWnd, $msg, $wParam, $lParam)
    If BitAND($wParam, 0xFFF0) <> 0xF180 Then Return $GUI_RUNDEFMSG ; SC_CONTEXTHELP
    _HelpDialog($hWnd, _HelpText())
    Return 1 ; we handled this click
EndFunc

; ==================================================================================================
; Helpers
; ==================================================================================================
; when using standard GUICtrlSetState($hEdit, $GUI_FOCUS), cursor position is
; always resets to line start (bug?), so use API function
Func _SetFocus($hWnd)
    DllCall("user32.dll", "hwnd", "SetFocus", "hwnd", $hWnd)
EndFunc

Func _GetFocus()
    Return DllCall("user32.dll", "hwnd", "GetFocus")[0]
EndFunc

Func _GUICtrlSetCustomFont($nCtrlHandle, $sConfigAlias)
    Local Static $aFontDefaults = [9, 400, 0, "MS Shell Dlg"]
    Local $sFont = StringStripWS(IniRead($g_sScriptConfig, "Appearance", $sConfigAlias, ""), 3)
    If $sFont = "" Then Return ; no any settings changed for this alias
    Local $aFont = StringSplit($sFont, ",", 2)
    If UBound($aFont) <> UBound($aFontDefaults) Then ReDim $aFont[UBound($aFontDefaults)]
    For $i = 0 To UBound($aFont) - 1
        If StringStripWS($aFont[$i], 3) = "" Then $aFont[$i] = $aFontDefaults[$i]
    Next
    GUICtrlSetFont($nCtrlHandle, $aFont[0], $aFont[1], $aFont[2], $aFont[3])
EndFunc

; from GUIListView.au3, simplified version, for control ID's only
Func __GUICtrlListView_SetItemSelected($hWnd, $iIndex, $bSelected = True, $bFocused = False)
    Local Const $tagLVITEM = "struct;uint Mask;int Item;int SubItem;uint State;uint StateMask;ptr Text;int TextMax;int Image;lparam Param;" & _
                             "int Indent;int GroupID;uint Columns;ptr pColumns;ptr piColFmt;int iGroup;endstruct"

	Local $tStruct = DllStructCreate($tagLVITEM)
	Local $iRet, $iSelected = 0, $iFocused = 0
	If ($bSelected = True) Then $iSelected = $LVIS_SELECTED
	If ($bFocused = True And $iIndex <> -1) Then $iFocused = $LVIS_FOCUSED
	DllStructSetData($tStruct, "Mask", $LVIF_STATE)
	DllStructSetData($tStruct, "Item", $iIndex)
	DllStructSetData($tStruct, "State", BitOR($iSelected, $iFocused))
	DllStructSetData($tStruct, "StateMask", BitOR($LVIS_SELECTED, $iFocused))
    $iRet = GUICtrlSendMsg($hWnd, $LVM_SETITEMSTATE, $iIndex, DllStructGetPtr($tStruct))
	Return $iRet <> 0
EndFunc

; Wrapper function for GUISetAccelerators to bypass issue when accelerators are 
; not working in any non-english layout (Autoit issue). GUISetAccelerators fails on setting shortcuts with english letters 
; in any layout, different from english. Though combinations with F-keys, Tab, Esc etc are working in any layout, accelerators 
; are set in one call, so they are affected too.
; Previous solution (CreatoR): load english layout when script starts. While it work fine on Win10, under Win7 it 
; set current layout to english at start.
; This solution: function change layout to english right before calling GUISetAccelerators and switch it back afterwards. 
; It's need to be used only on GUI, where shortcuts with englis letters used.
Func _GUISetAccelerators(ByRef $aAccelTable)
    Local Const $iEngLayout = 0x0409
    Static $bChecked = False, $hEngLayoutHandle = 0

    ; find if english layout loaded (performed once at first call)
    If Not $bChecked Then
        Local $aCall = DllCall('user32.dll', 'uint', 'GetKeyboardLayoutList', 'int', 0, 'ptr', 0)
        If Not @error And $aCall[0] Then
            Local $nCount = $aCall[0]
            Local $tData = DllStructCreate('handle[' & $nCount & ']')
            $aCall = DllCall('user32.dll', 'uint', 'GetKeyboardLayoutList', 'int', $aCall[0], 'struct*', $tData)
            If Not @error And $aCall[0] Then
                For $i = 1 To $nCount
                    ; it's not garanteed hkl-handle format will not be changed, so it's potentially unsafe code
                    If BitAND(DllStructGetData($tData, 1, $i), 0xFFFF) = $iEngLayout Then 
                        $hEngLayoutHandle = DllStructGetData($tData, 1, $i)
                        ExitLoop
                    EndIf
                Next
            EndIf
        EndIf
        $bChecked = True
    EndIf

    ; check current layout, if this is already english, do nothing, else process hack
    Local $bResult
    Local $aCall = DllCall("kernel32.dll", "long", "GetCurrentProcessId")
    $aCall = DllCall("user32.dll", "long", "GetKeyboardLayout", "long", $aCall[0])
    If BitAND($aCall[0], 0xFFFF) = $iEngLayout Then 
        $bResult = GUISetAccelerators($aAccelTable)
    Else
        ; load english layout if not loaded (unlikely though) or use already loaded
        Local $hEngLayout = $hEngLayoutHandle ? $hEngLayoutHandle : DllCall('user32.dll', 'handle', 'LoadKeyboardLayoutW', 'wstr', Hex($iEngLayout, 8), 'uint', 0)[0]
        ; switch to newly loaded layout
        Local $hPrevLayout = DllCall('user32.dll', 'handle', 'ActivateKeyboardLayout', 'handle', $hEngLayout, 'uint', 0)[0]
        ; load that damn accelerators
        $bResult = GUISetAccelerators($aAccelTable)
        ; switch layout back
        DllCall('user32.dll', 'handle', 'ActivateKeyboardLayout', 'handle', $hPrevLayout, 'uint', 0)
        ; unload layout (it's required or it will be shown in layouts list)
        If Not $hEngLayoutHandle Then DllCall('user32.dll', 'bool', 'UnloadKeyboardLayout', 'handle', $hEngLayout)
    EndIf 

    Return $bResult
EndFunc


; Simplified (only strings) sorting functions from Array.au3
Func _ArrayQuickSort1D(ByRef $aArray, Const ByRef $iStart, Const ByRef $iEnd)
    If $iEnd <= $iStart Then Return

	Local $vTmp, $L = $iStart, $R = $iEnd, $vPivot = $aArray[Int(($iStart + $iEnd) / 2)]
	Do
        While (StringCompare($aArray[$L], $vPivot) < 0)
            $L += 1
        WEnd
        While (StringCompare($aArray[$R], $vPivot) > 0)
            $R -= 1
        WEnd

		; Swap
		If $L <= $R Then
			$vTmp = $aArray[$L]
			$aArray[$L] = $aArray[$R]
			$aArray[$R] = $vTmp
			$L += 1
			$R -= 1
		EndIf
	Until $L > $R

	_ArrayQuickSort1D($aArray, $iStart, $R)
	_ArrayQuickSort1D($aArray, $L, $iEnd)
EndFunc

Func _ArrayQuickSort2D(ByRef $aArray, Const ByRef $iStep, Const ByRef $iStart, Const ByRef $iEnd, Const ByRef $iSubItem, Const ByRef $iSubMax)
	If $iEnd <= $iStart Then Return

	; QuickSort
	Local $vTmp, $L = $iStart, $R = $iEnd, $vPivot = $aArray[Int(($iStart + $iEnd) / 2)][$iSubItem]
	Do
			While ($iStep * StringCompare($aArray[$L][$iSubItem], $vPivot) < 0)
				$L += 1
			WEnd
			While ($iStep * StringCompare($aArray[$R][$iSubItem], $vPivot) > 0)
				$R -= 1
			WEnd

		; Swap
		If $L <= $R Then
			For $i = 0 To $iSubMax
				$vTmp = $aArray[$L][$i]
				$aArray[$L][$i] = $aArray[$R][$i]
				$aArray[$R][$i] = $vTmp
			Next
			$L += 1
			$R -= 1
		EndIf
	Until $L > $R

	_ArrayQuickSort2D($aArray, $iStep, $iStart, $R, $iSubItem, $iSubMax)
	_ArrayQuickSort2D($aArray, $iStep, $L, $iEnd, $iSubItem, $iSubMax)
EndFunc

Func _FileListToString(ByRef $aList)
    Local $sText = StringFormat($LANG("2"), UBound($aList)) & $cSep ; "|"
    For $i = 0 To UBound($aList)-1
        $sText &= $aList[$i] & $cSep ; "|"
    Next
    Return StringTrimRight($sText, 1)
EndFunc

Func _TryGetBasePath()
    ; this code is almost ported from this script:
    ; http://powerpro.wincmd.ru/Scripts/EditFileComments.htm
    Local $nPanel     = _TCHRequestInfo(1000)
    Local $nCaretItem = _TCHRequestInfo(1006 + $nPanel)
    Local $nUpdirItem = _TCHRequestInfo(1008 + $nPanel)
    ; if panel has no updir or caret is not on updir then error
    If ($nUpdirItem = 0) Or ($nCaretItem <> 0) Then Return SetError(1)
    Local $hPathPanel = _TCHRequestInfo(8 + $nPanel)
    Return _TCHExtractFilePath(ControlGetText($TC_HWND, "", $hPathPanel))
EndFunc

; ==================================================================================================
; Resources
; ==================================================================================================
Func Resource_Icon_Single_16()
    Return EIDecode(16, 16, "4LCgKAAAABAAMCAAMMIBAlAAAEAEAjA9AKoTAAa8AAbeAAY+LHaqDwADzAADhwADTAADqssI" & _
    "G4EAA/gAA/8QA6rBAAMWAAPEAAORAANqBgADsAADlQg/ATsRqxR3gRe3gAGfgAEHgAF6r4ABtYQFBQCBMxEABdWAAbOAAauA" & _
    "AQmUHZkf6m6AAbmAAQqEWYEBkVPrlR+BF56AAbuEBYFlgYf1oR8CwADcxA3BIcEO1Qr90QQBxBDhCtUvEQD/D/8PfQ0A+diP" & _
    "CQDBCMGaDQCD/8QQ0a/JAMEIwbo/AD8AHwABHAA=")
EndFunc

Func Resource_Icon_Multi_16()
    Return EIDecode(16, 16, "+bCgKAAAABAAMCAAMMIBAlAAAEAEAjA9AKoTAAa8AAbeAAY+LHaqDwADzAADhwADTAADqssQ" & _
    "I4EAA/gAA/8IA6rBAAMWAAPEAAORAANqBgADsAADlRA/ATsRqwwTgRe3gAGfgAEHgAG6r4ABtYQFkTsJAAWAAeqzgAGrgAEJ" & _
    "lB2JO40f6m6AAbmAAQqEWYEBiU/XjTuNH4EXnoABu4QFgWWrgYeJOxSUHwLAANzEDd/BIcEOzQjNDckGAcQQDQD35Q8VANkP" & _
    "+dBvCQDBBsE8/dUPg8QQyY/JAMEGwVzND//tHc0P7R0FAMEFwdTZHcEIOxUA4R2I/Ad/AHQA")
EndFunc
